/************************************************************************/
/* Symbol.cc: SpecC Internal Representation, Symbol Classes		*/
/************************************************************************/
/* Author: Rainer Doemer			first version: 10/08/97 */
/************************************************************************/

/* last update: 05/31/01 */

/* modifications: (most recent first)
 *
 * 05/31/01 RD	eliminated level 2 of SIR API
 * 05/25/01 RD	eliminated support for binary SIR files (import/export)
 * 05/25/01 RD	eliminated support for automatic IP wrapping (scc -ip option)
 * 05/23/01 RD	added implicit initialization with zero for class members
 *		which are not explicitly initialized by the user
 *		(see SIR_STORE_MAIN_BEHAVIOR_AS_STATIC)
 * 05/19/01 RD	fixed condition under which 'extern "C"' is generated
 * 04/30/01 RD	replaced use of obsolete form() from "stream.h" with own one
 * 04/27/01 RD	inserted ConvertInitializer() to perform explicit type
 *		conversion to expected type when entering into SIR
 *		(this avoids cumbersome conversion later);
 *		took out explicit type conversion during C++ output
 * 04/24/01 RD	adjusted explicit type conversion for initializers
 * 04/20/01 RD	added explicit type conversion for initializers
 * 02/14/01 RD	accomodate traversal routines for new constant node in
 *		port mapping tree (new constant/open mappings); adjusted
 *		C++ code generation
 * 01/30/01 RD	fixed a potential FMR problem in iterators
 *		SIR_Symbols::DFS_ForAllXxxxx
 * RD 06/27/00	added code for 12 and 16 byte alignment
 * RD 06/21/00	bug fix: another false warning about hidden symbols disabled
 * RD 06/21/00	bug fix: self-referencing enums crash (see DataTypes2.sc)
 * RD 06/18/00	Disabled warnings about function-pointer arguments
 *		hiding members of user-defined types (see DataTypes2.sc)
 * AG 05/26/00	Changed concatenation operator in port mappings to '@'
 */


#include "IntRep/Symbol.h"
#include "IntRep/Declarator.h"
#include "IntRep/Extern.h"
#include "Alignment.h"

#include <assert.h>


/*** constants and macros ***********************************************/


	/* optional code */

#define SIR_WARN_ABOUT_SHADOWED_SYMBOLS		/* during parsing only */


	/* location of the instantiation hierarchy */

/* Note:
 * SpecC LRM 1.0 requires that data at class level
 * is initialized with zero unless explicitly initialized
 * otherwise by the user. This requirement is automatically met
 * if the instantiation tree is allocated in static memory
 * (C++ guarantees zero-initialization of static data).
 * Unless there is other special initialization (not in the
 * current version), the following definition must be set
 * to SIR_STORE_MAIN_BEHAVIOR_AS_STATIC.
 */
//#define SIR_STORE_MAIN_BEHAVIOR_ON_STACK	/* instantiate as local var. */
#define SIR_STORE_MAIN_BEHAVIOR_AS_STATIC	/* instantiate as global var. */


/*** internal type declarations *****************************************/


	/* (none) */


/*** internal function prototypes ***************************************/


static void MarkImplementedMethods(	/* mark all methods with this name */
	const char	*Name,
	sir_symbol_ptr	*InterfaceList);


/*** class implementations **********************************************/


	/******************/
	/*** SIR_Symbol ***/
	/******************/


//++++++++++++++++++++++++++++ API Layer 1 +++++++++++++++++++++++++++++//


SIR_Symbol::SIR_Symbol(			/* constructor #1 (general) */
	SIR_SYMBOLCLASS	Class,
	const char	*Name,
	sir_type	*Type = NULL,
	SIR_STORAGE	StorageClass = SIR_STORAGE_NONE,
	int		PipeStages = 0,
	sir_initializer	*Initializer = NULL,
	unsigned int	Line = 0,
	sir_fileinfo	*FileInfo = NULL) :
		SIR_Node(Line, FileInfo)
{

assert(Name != NULL);
assert(*Name != 0);

SIR_Symbol::Class		= Class;
SIR_Symbol::Name		= Name;
SIR_Symbol::StorageClass	= StorageClass;
SIR_Symbol::PipeStages		= PipeStages;
SIR_Symbol::Type		= Type;
SIR_Symbol::Notes		= NULL;
SIR_Symbol::Initializer		= Initializer;
SIR_Symbol::EnumValue		= 0;
SIR_Symbol::Parameters		= NULL;
SIR_Symbol::ParamScope		= NULL;
SIR_Symbol::ClassScope		= NULL;
SIR_Symbol::Labels		= NULL;
SIR_Symbol::FctBody		= NULL;
SIR_Symbol::Interfaces		= NULL;
SIR_Symbol::PortMappings	= NULL;
SIR_Symbol::Imported		= NULL;
SIR_Symbol::Alias		= NULL;

} /* end of SIR_Symbol::SIR_Symbol #1 */


SIR_Symbol::SIR_Symbol(			/* constructor #2 (functions) */
	const char	*Name,
	sir_type	*Type,
	SIR_STORAGE	StorageClass,
	sir_parameters	*Parameters,
	sir_symbols	*ParamScope,
	sir_labels	*Labels,
	sir_statement	*FctBody,
	unsigned int	Line = 0,
	sir_fileinfo	*FileInfo = NULL) :
		SIR_Node(Line, FileInfo)
{

assert(Name != NULL);
assert(*Name != 0);

SIR_Symbol::Class		= SIR_SYMBOL_IDENTIFIER;
SIR_Symbol::Name		= Name;
SIR_Symbol::StorageClass	= StorageClass;
SIR_Symbol::PipeStages		= 0;
SIR_Symbol::Type		= Type;
SIR_Symbol::Notes		= NULL;
SIR_Symbol::Initializer		= NULL;
SIR_Symbol::EnumValue		= 0;
SIR_Symbol::Parameters		= Parameters;
SIR_Symbol::ParamScope		= ParamScope;
SIR_Symbol::ClassScope		= NULL;
SIR_Symbol::Labels		= Labels;
SIR_Symbol::FctBody		= FctBody;
SIR_Symbol::Interfaces		= NULL;
SIR_Symbol::PortMappings	= NULL;
SIR_Symbol::Imported		= NULL;
SIR_Symbol::Alias		= NULL;

if (ParamScope)
   { ParamScope->ParentSymbol = this;	/* automatically fill in back pointer */
    } /* fi */

} /* end of SIR_Symbol::SIR_Symbol #2 */


SIR_Symbol::SIR_Symbol(			/* constructor #3 (classes) */
	SIR_SYMBOLCLASS	Class,
	const char	*Name,
	sir_type	*Type,
	sir_parameters	*Parameters,
	sir_symbols	*ClassScope,
	sir_symbol_ptrs	*Interfaces,
	unsigned int	Line = 0,
	sir_fileinfo	*FileInfo = NULL) :
		SIR_Node(Line, FileInfo)
{

assert(Name != NULL);
assert(*Name != 0);

SIR_Symbol::Class		= Class;
SIR_Symbol::Name		= Name;
SIR_Symbol::StorageClass	= SIR_STORAGE_NONE;
SIR_Symbol::PipeStages		= 0;
SIR_Symbol::Type		= Type;
SIR_Symbol::Notes		= NULL;
SIR_Symbol::Initializer		= NULL;
SIR_Symbol::EnumValue		= 0;
SIR_Symbol::Parameters		= Parameters;
SIR_Symbol::ParamScope		= NULL;
SIR_Symbol::ClassScope		= ClassScope;
SIR_Symbol::Labels		= NULL;
SIR_Symbol::FctBody		= NULL;
SIR_Symbol::Interfaces		= Interfaces;
SIR_Symbol::PortMappings	= NULL;
SIR_Symbol::Imported		= NULL;
SIR_Symbol::Alias		= NULL;

if (ClassScope)
   { ClassScope->ParentSymbol = this;	/* automatically fill in back pointer */
    } /* fi */

} /* end of SIR_Symbol::SIR_Symbol #3 */


SIR_Symbol::SIR_Symbol(			/* constructor #4 (instantiations) */
	const char	*Name,
	sir_type	*Type,
	sir_portmaps	*PortMappings,
	unsigned int	Line = 0,
	sir_fileinfo	*FileInfo = NULL) :
		SIR_Node(Line, FileInfo)
{

assert(Name != NULL);
assert(Type != NULL);
assert(  (Type->Type == SIR_TYPE_BEHAVIOR)
       ||(Type->Type == SIR_TYPE_CHANNEL));
assert(PortMappings != NULL);

SIR_Symbol::Class		= SIR_SYMBOL_IDENTIFIER;
SIR_Symbol::Name		= Name;
SIR_Symbol::StorageClass	= SIR_STORAGE_NONE;
SIR_Symbol::PipeStages		= 0;
SIR_Symbol::Type		= Type;
SIR_Symbol::Notes		= NULL;
SIR_Symbol::Initializer		= NULL;
SIR_Symbol::EnumValue		= 0;
SIR_Symbol::Parameters		= NULL;
SIR_Symbol::ParamScope		= NULL;
SIR_Symbol::ClassScope		= NULL;
SIR_Symbol::Labels		= NULL;
SIR_Symbol::FctBody		= NULL;
SIR_Symbol::Interfaces		= NULL;
SIR_Symbol::PortMappings	= PortMappings;
SIR_Symbol::Imported		= NULL;
SIR_Symbol::Alias		= NULL;

} /* end of SIR_Symbol::SIR_Symbol #4 */


SIR_Symbol::SIR_Symbol(			/* constructor #6 (special f. parser) */
	sir_declspec	*DeclSpec,
	sir_declarator	*Declarator,
	sir_type	*Type,
	BOOL		StoreParameters = false) :
		SIR_Node(Declarator->Line, Declarator->FileInfo)
{

assert(DeclSpec != NULL);
assert(Declarator != NULL);
assert(Declarator->SymbolName != NULL);
assert(Type != NULL);

SIR_Symbol::Class		= (DeclSpec->StorageClass ==
						SIR_STORAGE_TYPEDEF) ?
					SIR_SYMBOL_TYPEDEF :
					SIR_SYMBOL_IDENTIFIER;
SIR_Symbol::Name		= Declarator->SymbolName->chars();
SIR_Symbol::StorageClass	= DeclSpec->StorageClass;
SIR_Symbol::PipeStages		= DeclSpec->PipeStages;
SIR_Symbol::Type		= Type;
SIR_Symbol::Notes		= NULL;
SIR_Symbol::Initializer		= NULL;
SIR_Symbol::EnumValue		= 0;
if (StoreParameters)
   { assert(Declarator->Parameters != NULL);
     assert(Declarator->ParamScope != NULL);
     SIR_Symbol::Parameters	= Declarator->Parameters;
     SIR_Symbol::ParamScope	= Declarator->ParamScope;
     Declarator->Parameters = NULL;
     Declarator->ParamScope = NULL;
     ParamScope->ParentSymbol = this;	/* automatically fill in back pointer */
    } /* fi */
else
   { SIR_Symbol::Parameters	= NULL;
     SIR_Symbol::ParamScope	= NULL;
    } /* esle */
SIR_Symbol::ClassScope		= NULL;
if (StoreParameters)
   { SIR_Symbol::Labels		= new SIR_Labels();	/* create label list */
    } /* fi */
else
   { SIR_Symbol::Labels		= NULL;
    } /* esle */
SIR_Symbol::FctBody		= NULL;
SIR_Symbol::Interfaces		= NULL;
SIR_Symbol::PortMappings	= NULL;
SIR_Symbol::Imported		= NULL;
SIR_Symbol::Alias		= NULL;

} /* end of SIR_Symbol::SIR_Symbol #6 */


SIR_Symbol::SIR_Symbol(			/* constructor #7 (special f. parser) */
	SIR_SYMBOLCLASS	Class,
	sir_declarator	*Declarator,
	sir_type	*Type,
	sir_symbol_ptrs	*Interfaces,
	BOOL		StoreParameters = false) :
		SIR_Node(Declarator->Line, Declarator->FileInfo)
{

assert(  (Class == SIR_SYMBOL_BEHAVIOR)
       ||(Class == SIR_SYMBOL_CHANNEL)
       ||(Class == SIR_SYMBOL_INTERFACE));
assert(Declarator != NULL);
assert(Declarator->SymbolName != NULL);
assert(Type != NULL);

SIR_Symbol::Class		= Class;
SIR_Symbol::Name		= Declarator->SymbolName->chars();
SIR_Symbol::StorageClass	= SIR_STORAGE_NONE;
SIR_Symbol::PipeStages		= 0;
SIR_Symbol::Type		= Type;
SIR_Symbol::Notes		= NULL;
SIR_Symbol::Initializer		= NULL;
SIR_Symbol::EnumValue		= 0;
if (StoreParameters)
   { /* Declarator->Parameters is NULL for interfaces */
     assert(Declarator->ParamScope != NULL);
     SIR_Symbol::Parameters	= Declarator->Parameters;
     SIR_Symbol::ClassScope	= Declarator->ParamScope;
     Declarator->Parameters = NULL;
     Declarator->ParamScope = NULL;
     SIR_Symbol::ClassScope->ParentSymbol = this; /* fill in back pointer */
    } /* fi */
else
   { SIR_Symbol::Parameters	= NULL;
     SIR_Symbol::ClassScope	= NULL;
    } /* esle */
SIR_Symbol::ParamScope		= NULL;
SIR_Symbol::Labels		= NULL;
SIR_Symbol::FctBody		= NULL;
SIR_Symbol::Interfaces		= Interfaces;
SIR_Symbol::PortMappings	= NULL;
SIR_Symbol::Imported		= NULL;
SIR_Symbol::Alias		= NULL;

} /* end of SIR_Symbol::SIR_Symbol #7 */


SIR_Symbol::SIR_Symbol(			/* constructor #8 (duplicator) */
	sir_symbol	*Original) :
		SIR_Node(Original)
{

SIR_Symbol::Class		= Original->Class;
SIR_Symbol::Name		= Original->Name;
SIR_Symbol::StorageClass	= Original->StorageClass;
SIR_Symbol::PipeStages		= Original->PipeStages;
SIR_Symbol::Type		= Original->Type;
SIR_Symbol::Notes		= Original->Notes ?
					new SIR_Notes(
						Original->Notes) :
					NULL;
SIR_Symbol::Initializer		= Original->Initializer ?
					new SIR_Initializer(
						Original->Initializer) :
					NULL;
SIR_Symbol::EnumValue		= Original->EnumValue;
SIR_Symbol::Parameters		= Original->Parameters ?
					new SIR_Parameters(
						Original->Parameters) :
					NULL;
SIR_Symbol::ParamScope		= Original->ParamScope ?
					new SIR_Symbols(
						Original->ParamScope) :
					NULL;
SIR_Symbol::ClassScope		= Original->ClassScope ?
					new SIR_Symbols(
						Original->ClassScope) :
					NULL;
SIR_Symbol::Labels		= Original->Labels ?
					new SIR_Labels(
						Original->Labels) :
					NULL;
SIR_Symbol::FctBody		= Original->FctBody ?
					new SIR_Statement(
						Original->FctBody) :
					NULL;
SIR_Symbol::Interfaces		= Original->Interfaces ?
					new SIR_SymbolPtrs(
						Original->Interfaces) :
					NULL;
SIR_Symbol::PortMappings	= Original->PortMappings ?
					new SIR_PortMaps(
						Original->PortMappings) :
					NULL;
SIR_Symbol::Imported		= Original->Imported;
SIR_Symbol::Alias		= NULL;

} /* end of SIR_Symbol::SIR_Symbol #8 */


SIR_Symbol::~SIR_Symbol(void)		/* destructor */
{

delete Notes;
delete Initializer;
delete Parameters;
delete ParamScope;
delete ClassScope;
delete Labels;
delete FctBody;
delete Interfaces;
delete PortMappings;

} /* end of SIR_Symbol::~SIR_Symbol */


BOOL SIR_Symbol::IsVariable(void)		/* check if IDENTIFIER */
{						/* and !TYPE_FUNCTION  */
						/* and !TYPE_BEHAVIOR  */
return(  (Class == SIR_SYMBOL_IDENTIFIER)	/* and !TYPE_CHANNEL   */
       &&(Type->Type != SIR_TYPE_FUNCTION)	/* and !TYPE_INTERFACE */
       &&(Type->Type != SIR_TYPE_BEHAVIOR)	/* and Direction==NONE */
       &&(Type->Type != SIR_TYPE_CHANNEL)	/* and !IsEnumMember   */
       &&(Type->Type != SIR_TYPE_INTERFACE)
       &&(Type->Direction == SIR_PORT_NONE)
       &&(!(IsEnumMember())));

} /* end of SIR_Symbol::IsVariable */


BOOL SIR_Symbol::IsVariableDefinition(void)	/* check if IsVariable */
{						/* and !STORAGE_EXTERN */

return(  (IsVariable())
       &&(StorageClass != SIR_STORAGE_EXTERN));

} /* end of SIR_Symbol::IsVariableDefinition */


BOOL SIR_Symbol::IsFunction(void)		/* check if IDENTIFIER */
{						/* and TYPE_FUNCTION   */

return(  (Class == SIR_SYMBOL_IDENTIFIER)
       &&(Type->Type == SIR_TYPE_FUNCTION));

} /* end of SIR_Symbol::IsFunction */


BOOL SIR_Symbol::IsFunctionDefinition(void)	/* check if IsFunction */
{						/* and FctBody!=NULL   */

return(  (IsFunction())
       &&(FctBody != NULL));

} /* end of SIR_Symbol::IsFunctionDefinition */


BOOL SIR_Symbol::IsBehaviorInstance(void)	/* check if IDENTIFIER */
{						/* and TYPE_BEHAVIOR   */
BOOL	Result;

Result = (  (Class == SIR_SYMBOL_IDENTIFIER)
	  &&(Type->Type == SIR_TYPE_BEHAVIOR));

if (Result)
   { assert(PortMappings != NULL);	/* no ports of type behavior */
    } /* fi */				/* (see below for channels) */

return(Result);

} /* end of SIR_Symbol::IsBehaviorInstance */


BOOL SIR_Symbol::IsChannelInstance(void)	/* check if IDENTIFIER */
{						/* and TYPE_CHANNEL    */

return(  (Class == SIR_SYMBOL_IDENTIFIER)
       &&(Type->Type == SIR_TYPE_CHANNEL));

} /* end of SIR_Symbol::IsChannelInstance */


BOOL SIR_Symbol::IsTypeDef(void)		/* check if TYPEDEF */
{

return(Class == SIR_SYMBOL_TYPEDEF);

} /* end of SIR_Symbol::IsTypeDef */


BOOL SIR_Symbol::IsBehavior(void)		/* check if BEHAVIOR */
{

return(Class == SIR_SYMBOL_BEHAVIOR);

} /* end of SIR_Symbol::IsBehavior */


BOOL SIR_Symbol::IsBehaviorDefinition(void)	/* check if IsBehavior  */
{						/* and ClassScope!=NULL */

return(  (IsBehavior())
       &&(ClassScope != NULL));

} /* end of SIR_Symbol::IsBehaviorDefinition */


BOOL SIR_Symbol::IsChannel(void)		/* check if CHANNEL */
{

return(Class == SIR_SYMBOL_CHANNEL);

} /* end of SIR_Symbol::IsChannel */


BOOL SIR_Symbol::IsChannelDefinition(void)	/* check if IsChannel   */
{						/* and ClassScope!=NULL */

return(  (IsChannel())
       &&(ClassScope != NULL));

} /* end of SIR_Symbol::IsChannelDefinition */


BOOL SIR_Symbol::IsInterface(void)		/* check if INTERFACE */
{

return(Class == SIR_SYMBOL_INTERFACE);

} /* end of SIR_Symbol::IsInterface */


BOOL SIR_Symbol::IsInterfaceDefinition(void)	/* check if IsInterface */
{						/* and ClassScope!=NULL */

return(  (IsInterface())
       &&(ClassScope != NULL));

} /* end of SIR_Symbol::IsInterfaceDefinition */


BOOL SIR_Symbol::IsClass(void)			/* check if BEHAVIOR or */
{						/* CHANNEL or INTERFACE */

return(  (Class == SIR_SYMBOL_BEHAVIOR)
       ||(Class == SIR_SYMBOL_CHANNEL)
       ||(Class == SIR_SYMBOL_INTERFACE));

} /* end of SIR_Symbol::IsClass */


BOOL SIR_Symbol::IsPort(void)			/* check if IDENTIFIER and */
{						/* TYPE_INTERFACE or       */
						/* Direction != PORT_NONE  */
return(  (Class == SIR_SYMBOL_IDENTIFIER)
       &&(Type->IsPort()));

} /* end of SIR_Symbol::IsPort */


BOOL SIR_Symbol::IsParameter(void)		/* check if defined in */
{						/* SCOPE_PARAMETER */

return(GetScope()->ScopeInfo == SIR_SCOPE_PARAMETER);

} /* end of SIR_Symbol::IsParameter */


BOOL SIR_Symbol::IsEnumMember(void)		/* check if this is an enum */
{						/* member (a constant) */

return(  (Class == SIR_SYMBOL_IDENTIFIER)
       &&(Type->Type == SIR_TYPE_ENUM)
       &&(  (EnumValue != 0)			// variables don't have value
	  ||(Type->UserType->Members == NULL)	// still under construction
	  ||(Type->UserType->Members->Find(this))));	// inefficient, but
							// the only way with
							// this data structure

} /* end of SIR_Symbol::IsEnumMember */


BOOL SIR_Symbol::IsDirectDependant(	/* checks if this depends on that */
	sir_symbol	*ThatSymbol,	/* (directly) */
	SIR_DEPENDENCY	*Reason = NULL)
{
sir_symbol_ptr	*Interface;
sir_portmap	*PortMap;
sir_bitslice	*BitSlice;

assert(ThatSymbol != NULL);
assert(this != ThatSymbol);	/* nothing depends on itself */
assert(Type != NULL);

if (  (Type->ClassSymbol)
    &&(Type->ClassSymbol == ThatSymbol))
   { assert(  (IsBehaviorInstance())
	    ||(IsChannelInstance()));
     assert(ThatSymbol->IsClass());
     if (Reason)
	{ *Reason = SIR_DEP_INSTANTIATION;	/* is an inst. of that class */
	 } /* fi */
     return(TRUE);
    } /* fi */

/* FctBody may contain _indirect_ dependency */

if (Interfaces)
   { Interface = Interfaces->First();
     while(Interface)
	{ if (Interface->Symbol == ThatSymbol)
	     { if (Reason)
		  { *Reason = SIR_DEP_IMPLEMENTATION;
		   } /* fi */
	       return(TRUE);		/* is implementing that interface */
	      } /* fi */
	  Interface = Interface->Succ();
	 } /* elihw */
    } /* fi */

if (PortMappings)
   { PortMap = PortMappings->First();
     while(PortMap)
	{ assert(PortMap->BitSlices != NULL);
	  BitSlice = PortMap->BitSlices->First();
	  while(BitSlice)
	     { if (BitSlice->Symbol == ThatSymbol)
		  { if (Reason)
		       { *Reason = SIR_DEP_CONNECTION;
			} /* fi */
		    return(TRUE);	/* is mapped to that symbol */
		   } /* fi */
	       BitSlice = BitSlice->Succ();
	      } /* elihw */
	  PortMap = PortMap->Succ();
	 } /* elihw */
    } /* fi */

return(FALSE);	/* is independent */

} /* end of SIR_Symbol::IsDirectDependant */


BOOL SIR_Symbol::IsIndirectDependant(	/* checks if this depends on that */
	sir_symbol	*ThatSymbol,	/* (indirectly) */
	SIR_DEPENDENCY	*Reason = NULL,
	sir_statement	**DepStmnt = NULL,
	sir_expression	**DepExpr = NULL)
{

assert(ThatSymbol != NULL);
assert(this != ThatSymbol);	/* nothing depends on itself */

if (FctBody)
   { if (FctBody->DFS_FindDependant(ThatSymbol,
				DepStmnt, DepExpr, Reason))
	{ return(TRUE);
	 } /* fi */
    } /* fi */

return(FALSE);	/* is independent */

} /* end of SIR_Symbol::IsIndirectDependant */


BOOL SIR_Symbol::MapsPortTo(		/* checks for port-instantiation */
	sir_symbol	*Symbol)	/* (instantiations only) */
{
sir_portmap	*PortMap;
sir_bitslice	*BitSlice;

assert(PortMappings != NULL);	/* must have a port map list */
PortMap = PortMappings->First();
while(PortMap)
   { assert(PortMap->BitSlices != NULL);	/* must have a mapping */
     BitSlice = PortMap->BitSlices->First();
     while(BitSlice)
	{ if (BitSlice->Symbol == Symbol)
	     { return(TRUE);		/* port instance found */
	      } /* fi */
	  BitSlice = BitSlice->Succ();
	 } /* elihw */
     PortMap = PortMap->Succ();
    } /* elihw */

return(FALSE);	/* not found */

} /* end of SIR_Symbol::MapsPortTo */


BOOL SIR_Symbol::Instantiates(		/* checks for instantiation */
	sir_symbol	*ClassSymbol)	/* (behaviors and channels only) */
{
sir_symbol	*Symbol;

if (! ClassScope)
   { return(FALSE);	/* declaration only (black box) */
    } /* fi */

Symbol = ClassScope->First();
while(Symbol)
   { if (  (  (Symbol->IsBehaviorInstance())
	    ||(Symbol->IsChannelInstance()))
	 &&(Symbol->Type->ClassSymbol == ClassSymbol))
	{ return(TRUE);		/* instantiation found */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

return(FALSE);	/* not found */

} /* end of SIR_Symbol::Instantiates */


BOOL SIR_Symbol::Contains(		/* checks for containment */
	sir_symbol	*ClassSymbol)	/* (behaviors and channels only) */
{
sir_symbol	*Symbol;

if (! ClassScope)
   { return(FALSE);	/* declaration only (black box) */
    } /* fi */

Symbol = ClassScope->First();
while(Symbol)
   { if (  (Symbol->IsBehaviorInstance())
	 ||(Symbol->IsChannelInstance()))
	{ if (Symbol->Type->ClassSymbol == ClassSymbol)
	     { return(TRUE);	/* instantiation found */
	      } /* fi */
	  if (Symbol->Type->ClassSymbol->Contains(ClassSymbol))	/* recursion! */
	     { return(TRUE);	/* nested instantiation found */
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

return(FALSE);	/* not found */

} /* end of SIR_Symbol::Contains */


sir_symbols *SIR_Symbol::GetScope(void)	/* determines scope of this symbol */
{					/* (returns NULL if not in any table) */

return((sir_symbols*) Head());

} /* end of SIR_Symbol::GetScope */


sir_design *SIR_Symbol::GetDesign(	/* determines the design */
	void)			/* (returns NULL if type is not in design) */
{
sir_types	*TypeTable;

if (  (Type)
    &&(TypeTable = Type->GetTable()))
   { return(TypeTable->ThisDesign);
    } /* fi */

return(NULL);

} /* end of SIR_Symbol::GetDesign */


ERROR SIR_Symbol::DFS_ForAllNodes(	/* iterator over all nodes */
	sir_node_mptr	MemberFct,	/* (depth first) */
	sir_node_marg	MemberFctArg)
{

if ((SIR_Error = (this->*MemberFct)(MemberFctArg)))	/* process this node */
   { return(SIR_Error);
    } /* fi */

if (Notes)
   { if ((SIR_Error = Notes->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (Initializer)
   { if ((SIR_Error = Initializer->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (ParamScope)
   { if ((SIR_Error = ParamScope->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (ClassScope)
   { if ((SIR_Error = ClassScope->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (Labels)
   { if ((SIR_Error = Labels->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (FctBody)
   { if ((SIR_Error = FctBody->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (PortMappings)
   { if ((SIR_Error = PortMappings->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::DFS_ForAllNodes */


ERROR SIR_Symbol::DFS_ForAllSymbols(	/* iterator over all symbols */
	sir_symbol_mptr	MemberFct,	/* (depth first) */
	sir_symbol_marg	MemberFctArg)
{

if ((SIR_Error = (this->*MemberFct)(MemberFctArg)))	/* process this node */
   { return(SIR_Error);
    } /* fi */

/* there are no symbols in Notes */
/* there are no symbols in Initializer */
/* there are no symbols in Parameters */
if (ParamScope)
   { if ((SIR_Error = ParamScope->DFS_ForAllSymbols(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (ClassScope)
   { if ((SIR_Error = ClassScope->DFS_ForAllSymbols(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
/* there are no symbols in Labels */
if (FctBody)
   { if ((SIR_Error = FctBody->DFS_ForAllSymbols(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
/* there are no symbols in Interfaces */
/* there are no symbols in PortMappings */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::DFS_ForAllSymbols */


ERROR SIR_Symbol::DFS_ForAllUserTypes(	/* iterator over all usertypes */
	sir_usertp_mptr	MemberFct,	/* (depth first) */
	sir_usertp_marg	MemberFctArg)
{

/* this is not a usertype */

/* there are no usertypes in Notes */
/* there are no usertypes in Initializer */
/* there are no usertypes in Parameters */
if (ParamScope)
   { if ((SIR_Error = ParamScope->DFS_ForAllUserTypes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (ClassScope)
   { if ((SIR_Error = ClassScope->DFS_ForAllUserTypes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
/* there are no usertypes in Labels */
if (FctBody)
   { if ((SIR_Error = FctBody->DFS_ForAllUserTypes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
/* there are no usertypes in Interfaces */
/* there are no usertypes in PortMappings */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::DFS_ForAllUserTypes */


ERROR SIR_Symbol::DFS_ForAllNotes(	/* iterator over all notes */
	sir_note_mptr	MemberFct,	/* (depth first) */
	sir_note_marg	MemberFctArg)
{

/* this is not a note */

if (Notes)
   { if ((SIR_Error = Notes->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
/* there are no notes in Initializer */
/* there are no notes in Parameters */
if (ParamScope)
   { if ((SIR_Error = ParamScope->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (ClassScope)
   { if ((SIR_Error = ClassScope->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (Labels)
   { if ((SIR_Error = Labels->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (FctBody)
   { if ((SIR_Error = FctBody->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
/* there are no notes in Interfaces */
/* there are no notes in PortMappings */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::DFS_ForAllNotes */


ERROR SIR_Symbol::SetImported(	/* sets the Imported pointer (if unset) */
	sir_symbol_marg	ImportPtr)	/* or resets it if NULL */
{

if (ImportPtr == NULL)
   { Imported = NULL;
     return(SIR_ERROR_NO_ERROR);
    } /* fi */

if (! Imported)		/* import pointer still unset? */
   { Imported = (sir_import*) ImportPtr;
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::SetImported */


ERROR SIR_Symbol::TakeOverImport(	/* takes over an imported symbol */
	sir_symbol_marg	ImportPtr)
{

assert(ImportPtr != NULL);
if (Imported == (sir_import*) ImportPtr)
   { Imported = NULL;
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::TakeOverImport */


ERROR SIR_Symbol::MarkImportEntry(	/* marks its import entry being used */
	sir_symbol_marg	/* Unused */)
{

if (Imported)
   { Imported->Color = SIR_RED;	/* mark as being used */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::MarkImportEntry */


ERROR SIR_Symbol::MarkUsedTypes( /* marks the type entries used here */
	sir_symbol_marg	/* Unused */)
{
sir_parameter	*Para;

Type->MarkUsed();

if (Parameters)
   { Para = Parameters->First();
     while(Para)
	{ Para->Type->MarkUsed();
	  Para = Para->Succ();
	 } /* elihw */
    } /* fi */

if (FctBody)
   { FctBody->DFS_ForAllExpressions(
		&SIR_Expression::MarkUsedTypes, NULL);
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::MarkUsedTypes */


BOOL SIR_Symbol::DFS_FindDependant(	/* searches for dependants (DFS) */
	sir_symbol	**SkipSymbol,	/* (internal helper function) */
	sir_symbol	*ThatSymbol,
	sir_symbol	**DepSymbol,
	sir_statement	**DepStmnt,
	sir_expression	**DepExpr,
	SIR_DEPENDENCY	*Reason)
{

assert(SkipSymbol != NULL);
assert(ThatSymbol != NULL);

if (*SkipSymbol)
   { if (this == *SkipSymbol)
	{ *SkipSymbol = NULL;	/* no more skipping */
	 } /* fi */
     return(FALSE);	/* skip this symbol */
    } /* fi */

if (IsDirectDependant(ThatSymbol, Reason))	/* check this symbol */
   { if (DepSymbol)				/* (for direct dep.) */
	{ *DepSymbol = this;
	 } /* fi */
     if (DepStmnt)
	{ *DepStmnt = NULL;
	 } /* fi */
     if (DepExpr)
	{ *DepExpr = NULL;
	 } /* fi */
     return(TRUE);
    } /* fi */

if (IsIndirectDependant(ThatSymbol, Reason,	/* check this symbol */
			DepStmnt, DepExpr))	/* (for indirect dep.) */
   { if (DepSymbol)
	{ *DepSymbol = this;
	 } /* fi */
     return(TRUE);
    } /* fi */

/* there are no symbols in Notes */
/* there are no symbols in Initializer */
/* there are no symbols in Parameters */
/* symbols in ParamScope cannot depend on other symbols */

if (ClassScope)
   { if (ClassScope->DFS_FindDependant(SkipSymbol, ThatSymbol,
					DepSymbol, DepStmnt, DepExpr, Reason))
	{ return(TRUE);
	 } /* fi */
    } /* fi */

/* there are no symbols in Labels */
/* symbols in FctBody cannot depend on other symbols */
/* there are no symbols in Interfaces */
/* there are no symbols in PortMappings */

return(FALSE);

} /* end of SIR_Symbol::DFS_FindDependant */


BOOL SIR_Symbol::FindDependant(	/* finds symbols that depend on this one #1 */
	sir_symbol	**DepSymbol = NULL,	/* return symbol found */
	sir_statement	**DepStmnt = NULL,	/* return statement found */
	sir_expression	**DepExpr = NULL,	/* return expression found */
	SIR_DEPENDENCY	*Reason = NULL,		/* return type of dependency */
	sir_symbols	*SearchScope = NULL,	/* search in this scope */
	sir_symbol	*SearchAfter = NULL)	/* continue search here */
{
sir_symbols	*RootScope;
sir_symbol	*SkipSymbol;

if (SearchScope)			/* determine where to search */
   { assert(SearchScope->ScopeInfo
		!= SIR_SCOPE_USERTYPE);	/* we don't go there! */
     RootScope = SearchScope;
    } /* fi */
else
   { RootScope = GetScope();
     while(RootScope->Parent)
	{ RootScope = RootScope->Parent;
	 } /* elihw */
    } /* esle */

if (SearchAfter)			/* determine starting point */
   { assert(SearchAfter->GetScope()->ScopeInfo
		!= SIR_SCOPE_USERTYPE);	/* we don't go there! */
     SkipSymbol = SearchAfter;
    } /* fi */
else
   { SkipSymbol = NULL;	/* don't skip */
    } /* esle */

if (RootScope->DFS_FindDependant(	/* recursive search */
		&SkipSymbol, this,
		DepSymbol, DepStmnt, DepExpr, Reason))
   { return(TRUE);	/* found one! */
    } /* fi */

/* no dependant found! */

if (DepSymbol)
   { *DepSymbol = NULL;
    } /* fi */
if (DepStmnt)
   { *DepStmnt = NULL;
    } /* fi */
if (DepExpr)
   { *DepExpr = NULL;
    } /* fi */
if (Reason)
   { *Reason = SIR_DEP_NONE;
    } /* fi */

return(FALSE);

} /* end of SIR_Symbol::FindDependant #1 */


BOOL SIR_Symbol::FindDependant(	/* finds symbols that depend on this one #2 */
	sir_statement	*SearchStmnt,		/* search in this statement */
	sir_statement	**DepStmnt = NULL,	/* return statement found */
	sir_expression	**DepExpr = NULL,	/* return expression found */
	SIR_DEPENDENCY	*Reason = NULL)		/* return type of dependency */
{

assert(SearchStmnt != NULL);

if (SearchStmnt->DFS_FindDependant(	/* recursive search */
		this, DepStmnt, DepExpr, Reason))
   { return(TRUE);	/* found one! */
    } /* fi */

/* no dependant found! */

if (DepStmnt)
   { *DepStmnt = NULL;
    } /* fi */
if (DepExpr)
   { *DepExpr = NULL;
    } /* fi */
if (Reason)
   { *Reason = SIR_DEP_NONE;
    } /* fi */

return(FALSE);

} /* end of SIR_Symbol::FindDependant #2 */


void SIR_Symbol::SetAlias(	/* sets all type, usertype, symbol alias' */
	sir_symbol	*Alias)	/* (iterates over symbols and usertypes) */
{

assert(Alias != NULL);

this->Alias = Alias;	/* process this node */

/* there are no symbols or usertypes in Notes */
/* there are no symbols or usertypes in Initializer */
/* there are no symbols or usertypes in Parameters */
if (ParamScope)
   { ParamScope->SetAlias(Alias->ParamScope);
    } /* fi */
if (ClassScope)
   { ClassScope->SetAlias(Alias->ClassScope);
    } /* fi */
/* there are no symbols or usertypes in Labels */
if (FctBody)
   { FctBody->SetAlias(Alias->FctBody);
    } /* fi */
/* there are no symbols or usertypes in Interfaces */
/* there are no symbols or usertypes in PortMappings */

} /* end of SIR_Symbol::SetAlias */


void SIR_Symbol::UnAlias(	/* unalias all type, usertype, symbol links */
	sir_symbols	*GlobalSymbols)
{

if (Type->Alias)
   { Type = Type->Alias;
    } /* fi */
/* there is nothing to unalias in Notes (it's built-in) */
/* there is nothing to unalias in Initializer */
if (Parameters)
   { Parameters->UnAlias();
    } /* fi */
if (ParamScope)
   { ParamScope->UnAlias(GlobalSymbols);
    } /* fi */
if (ClassScope)
   { ClassScope->UnAlias(GlobalSymbols);
    } /* fi */
/* there is nothing to unalias in Labels */
if (FctBody)
   { FctBody->UnAlias(GlobalSymbols);
    } /* fi */
if (Interfaces)
   { Interfaces->UnAlias();
    } /* fi */
if (PortMappings)
   { PortMappings->UnAlias();
    } /* fi */

} /* end of SIR_Symbol::UnAlias */


const char *SIR_Symbol::PrintStorageClass(	/* prints storage class #1 */
	BOOL		AppendSpace = false)  /* (eg. "extern" or "extern ") */
{

return(PrintStorageClass(StorageClass, PipeStages, AppendSpace));

} /* end of SIR_Symbol::PrintStorageClass */


const char *SIR_Symbol::PrintStorageClass(	/* prints storage class #2 */
	SIR_STORAGE	StorageClass,	      /* (eg. "extern" or "extern ") */
	int		PipeStages = 0,
	BOOL		AppendSpace = false)
{
static string	Buffer;
int		i;

switch(StorageClass)
   { case SIR_STORAGE_NONE:
	{ return("");
	 }
     case SIR_STORAGE_TYPEDEF:
	{ if (AppendSpace)
	     { return("typedef ");
	      } /* fi */
	  else
	     { return("typedef");
	      } /* esle */
	 }
     case SIR_STORAGE_EXTERN:
	{ if (AppendSpace)
	     { return("extern ");
	      } /* fi */
	  else
	     { return("extern");
	      } /* esle */
	 }
     case SIR_STORAGE_STATIC:
	{ if (AppendSpace)
	     { return("static ");
	      } /* fi */
	  else
	     { return("static");
	      } /* esle */
	 }
     case SIR_STORAGE_AUTO:
	{ if (AppendSpace)
	     { return("auto ");
	      } /* fi */
	  else
	     { return("auto");
	      } /* esle */
	 }
     case SIR_STORAGE_REGISTER:
	{ if (AppendSpace)
	     { return("register ");
	      } /* fi */
	  else
	     { return("register");
	      } /* esle */
	 }
     case SIR_STORAGE_PIPED:
	{ Buffer = "";
	  for(i=0; i<PipeStages-1; i++)
	     { Buffer += "piped ";
	      } /* rof */
	  if (AppendSpace)	/* even if PipeStages = default = 0 print one */
	     { Buffer += "piped ";
	      } /* fi */
	  else
	     { Buffer += "piped";
	      } /* esle */
	  break;
	 }
     default:
	{ assert(FALSE);	/* bad storage class */
	 }
    } /* hctiws */

return(Buffer.chars());

} /* end of SIR_Symbol::PrintStorageClass */


ERROR SIR_Symbol::WriteSC(	/* (re-) generates SpecC source code */
	FILE		*File,	/* (for non-class symbols) */
	BOOL		WriteNotes,
	BOOL		AsDeclaration = FALSE,
	BOOL		CplusplusMode = FALSE,
	BOOL		AsReference = FALSE)
{
string		Buffer;
BOOL		PutLineInfoHere = FALSE,
		PutNotesHere = FALSE;
BOOL		MakePureVirtual;

if (IsPort())	/* ports are written with the class */
   { return(SIR_ERROR_NO_ERROR);
    } /* fi */

switch(Class)
   { case SIR_SYMBOL_IDENTIFIER:
	{ switch(Type->Type)
	     { case SIR_TYPE_FUNCTION:
		  { assert(AsDeclaration == TRUE); /* fct.def. not supported */
		    if (IsFunctionDefinition())
		       { PutLineInfoHere	= FALSE;
			 PutNotesHere		= TRUE;
			} /* fi */
		    else
		       { PutLineInfoHere	= TRUE;
			 PutNotesHere		= TRUE;
			} /* esle */
		    break;
		   }
	       case SIR_TYPE_BEHAVIOR:
	       case SIR_TYPE_CHANNEL:
		  { assert(AsDeclaration == FALSE); /* no instance decl. */
		    PutLineInfoHere	= TRUE;
		    PutNotesHere	= TRUE;
		    break;
		   }
	       default:	/* any other TYPE (standard variables) */
		  { if (AsDeclaration)
		       { if (IsVariableDefinition())
			    { /* no declaration necessary */
			      return(SIR_ERROR_NO_ERROR);
			     } /* fi */
			 else
			    { PutLineInfoHere		= TRUE;
			      PutNotesHere		= TRUE;
			     } /* esle */
			} /* fi */
		    else
		       { if (IsVariableDefinition())
			    { PutLineInfoHere		= TRUE;
			      PutNotesHere		= TRUE;
			     } /* fi */
			 else
			    { /* is just a declaration */
			      return(SIR_ERROR_NO_ERROR);
			     } /* esle */
			} /* esle */
		    break;
		   }
	      } /* hctiws */
	  break;
	 }
     case SIR_SYMBOL_TYPEDEF:
	{ PutLineInfoHere	= TRUE;
	  PutNotesHere		= TRUE;
	  break;
	 }
     case SIR_SYMBOL_BEHAVIOR:
     case SIR_SYMBOL_CHANNEL:
     case SIR_SYMBOL_INTERFACE:
	{ assert(FALSE);	/* classes not supported */
 	 }
     default:
	{ assert(FALSE);	/* bad symbol class */
	 }
    } /* hctiws */

MakePureVirtual = (  (CplusplusMode)
		   &&(GetScope())
		   &&(GetScope()->ScopeInfo == SIR_SCOPE_CLASS)
		   &&(GetScope()->ParentSymbol->Class == SIR_SYMBOL_INTERFACE));

if (PutLineInfoHere)
   { if (LineInfo)
	{ if ((SIR_Error = LineInfo->WriteSC(File, true)))	/* old stuff */
	     { return(SIR_Error);				/* (by user) */
	      } /* fi */
	 } /* fi */
     else
	{ SIR_LineInfo::WriteNewSC(File, true);	/* new stuff (by tools) */
	 } /* esle */
    } /* fi */
else
   { SIR_LineInfo::WriteNewSC(File, true);	/* new or old stuff (decls) */
    } /* esle */

if (CplusplusMode)
   { if (  (  (StorageClass == SIR_STORAGE_EXTERN)
	    ||(StorageClass == SIR_STORAGE_NONE))
	 &&(IsFunction())
	 &&(FctBody==NULL)
	 &&(GetScope()->ScopeInfo == SIR_SCOPE_GLOBAL))
	{ assert(AsReference == FALSE);
	  fputs("extern \"C\" ", File);	/* C linkage! */
	  Buffer = Name;
	  fputs(Type->PrettyString(&Buffer, WriteNotes, CplusplusMode),
				File);
	 } /* fi */
     else
	{ if (StorageClass == SIR_STORAGE_PIPED)
	     { assert(AsReference == FALSE);
	       fputs(SIR_SIM_PIPED_NAME "<", File);
	       Buffer = "";
	       fputs(Type->PrettyString(&Buffer, WriteNotes, CplusplusMode),
				File);
	       assert(PipeStages >= 1);
	       fputs("> ", File);
	       fputs(Name.chars(), File);
	       if (GetScope()->ScopeInfo != SIR_SCOPE_CLASS)
		  { fprintf(File, "(%d)", PipeStages);
		   } /* fi */
	      } /* fi */
	  else
	     { fputs(PrintStorageClass(TRUE), File);
	       if (MakePureVirtual)
		  { fputs("virtual ", File);
		   } /* fi */
	       if (AsReference)
		  { Buffer = "(&";
		    Buffer += Name;
		    Buffer += ")";
		   } /* fi */
	       else
		  { Buffer = Name;
		   } /* esle */
	       fputs(Type->PrettyString(&Buffer, WriteNotes, CplusplusMode),
				File);
	      } /* fi */
	 } /* fi */
    } /* fi */
else
   { assert(AsReference == FALSE);
     fputs(PrintStorageClass(TRUE), File);
     Buffer = Name;
     fputs(Type->PrettyString(&Buffer, WriteNotes, CplusplusMode), File);
    } /* esle */

if (MakePureVirtual)
   { assert(Initializer == NULL); /* would collide */
     fputs(" = 0", File);
    } /* fi */

if (  (Initializer)
    &&(! AsDeclaration)
    &&(  (! CplusplusMode)
       ||((CplusplusMode)&&(GetScope()->ScopeInfo != SIR_SCOPE_CLASS))))
   { fputs(" = ", File);
     if ((SIR_Error = Initializer->WriteSC(File, CplusplusMode)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (  (! CplusplusMode)
    &&(  (IsBehaviorInstance())
       ||(IsChannelInstance())))
   { assert(PortMappings != NULL);	/* port mapping */
     PortMappings->WriteSC(File);
    } /* fi */
fputc(';', File);

if (ferror(File))
   { SIR_ErrMsg.form("Writing to file failed%s",
			GL_SystemErrorMessage());
     return(SIR_ERROR_WRITING_SPECC_FILE);
    } /* fi */

if (PutNotesHere)
   { if (Notes)
	{ if ((SIR_Error = Notes->WriteSC(File, WriteNotes, CplusplusMode)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::WriteSC */


ERROR SIR_Symbol::WriteSC2(	/* (re-) generates SpecC source code */
	FILE		*File,	/* (for class symbols) */
	BOOL		WriteNotes,
	BOOL		AsDeclaration = FALSE)
{
char		*ClassName = NULL;
string		Buffer;
sir_parameter	*Port;
sir_type_ptr	*PortType;
sir_symbol_ptr	*Interface;
sir_symbol	*ParamSymbol;

if (  ((AsDeclaration) && (ClassScope == NULL))
    ||((! AsDeclaration) && (ClassScope != NULL)))
   { if (LineInfo)
	{ if ((SIR_Error = LineInfo->WriteSC(File, true)))	/* old stuff */
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     else
	{ SIR_LineInfo::WriteNewSC(File, true);	/* new stuff */
	 } /* esle */
    } /* fi */
else
   { SIR_LineInfo::WriteNewSC(File, true);	/* old or new stuff */
    } /* esle */

switch(Class)
   { case SIR_SYMBOL_IDENTIFIER:
     case SIR_SYMBOL_TYPEDEF:
	{ assert(FALSE);	/* this is for classes only */
	 }
     case SIR_SYMBOL_BEHAVIOR:
	{ ClassName = "behavior ";
	  break;
	 }
     case SIR_SYMBOL_CHANNEL:
	{ ClassName = "channel ";
	  break;
	 }
     case SIR_SYMBOL_INTERFACE:
	{ ClassName = "interface ";
	  break;
	 }
     default:
	{ assert(FALSE);	/* bad symbol class */
	 }
    } /* hctiws */
fputs(ClassName, File);
fputs(Name.chars(), File);

if (Class != SIR_SYMBOL_INTERFACE)
   { fputs(" (", File);
     if (AsDeclaration)
	{ PortType = Type->Parameters->First();
	  while(PortType)
	     { Buffer = "";
	       fputs(PortType->Type->PrettyString(&Buffer, WriteNotes), File);
	       PortType = PortType->Succ();
	       if (PortType)
		  { fputs(", ", File);
		   } /* fi */
	      } /* elihw */
	 } /* fi */
     else
	{ assert(Parameters != NULL);
	  Port = Parameters->First();
	  while(Port)
	     { if (Port->Symbol)
		  { Buffer = Port->Symbol->Name;
		   } /* fi */
	       else
		  { Buffer = "";
		   } /* esle */
	       fputs(Port->Type->PrettyString(&Buffer, WriteNotes), File);
	       Port = Port->Succ();
	       if (Port)
		  { fputs(", ", File);
		   } /* fi */
	      } /* elihw */
	 } /* esle */
     fputc(')', File);
    } /* fi */

if (  (Interfaces)
    &&(! Interfaces->Empty()))
   { fputs(" implements ", File);
     Interface = Interfaces->First();
     while(Interface)
	{ fputs(Interface->Symbol->Name.chars(), File);
	  Interface = Interface->Succ();
	  if (Interface)
	     { fputs(", ", File);
	      } /* fi */
	 } /* elihw */
    } /* fi */

if (  (! AsDeclaration)
    &&(ClassScope))
   { SIR_LineInfo::WriteNL(File);
     fputc('{', File);
     SIR_LineInfo::Add2Tabulator(1);
     ParamSymbol = ClassScope->First();
     while(ParamSymbol)
	{ if (  (ParamSymbol->IsPort())
	      &&(ParamSymbol->Notes))
	     { if ((SIR_Error = ParamSymbol->Notes->WriteSC(File, WriteNotes)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  ParamSymbol = ParamSymbol->Succ();
	 } /* elihw */
     if ((SIR_Error = ClassScope->WriteSC(File, WriteNotes)))
	{ return(SIR_Error);
	 } /* fi */
     SIR_LineInfo::Add2Tabulator(-1);
     SIR_LineInfo::WriteNL(File);
     fputc('}', File);
    } /* fi */

fputc(';', File);

if (ferror(File))
   { SIR_ErrMsg.form("Writing to file failed%s",
			GL_SystemErrorMessage());
     return(SIR_ERROR_WRITING_SPECC_FILE);
    } /* fi */

if (AsDeclaration)
   { if (Notes)
	{ if ((SIR_Error = Notes->WriteSC(File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::WriteSC2 */


ERROR SIR_Symbol::WriteSC3(	/* (re-) generates SpecC source code */
	FILE		*File,	/* (for function definitions) */
	BOOL		WriteNotes,
	BOOL		CplusplusMode = FALSE)
{
string		Buffer,
		ParamBuffer;
sir_parameter	*Parameter;

assert(IsFunctionDefinition());	/* doesn't support anything else */

if (LineInfo)
   { if ((SIR_Error = LineInfo->WriteSC(File, true)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
else
   { SIR_LineInfo::WriteNewSC(File, true);
    } /* esle */

fputs(PrintStorageClass(true), File);
if (  (CplusplusMode)
    &&(GetScope()->ScopeInfo == SIR_SCOPE_CLASS))
   { Buffer = GetScope()->ParentSymbol->Name;
     Buffer += "::";
     Buffer += Name;
    } /* fi */
else
   { Buffer = Name;
    } /* fi */
Buffer += "(";
Parameter = Parameters->First();
while(Parameter)
   { if (Parameter->Symbol)
	{ Buffer += Parameter->Symbol->PrintStorageClass(TRUE);
	  ParamBuffer = Parameter->Symbol->Name;
	 } /* fi */
     else
	{ ParamBuffer = "";
	 } /* esle */
     if (  (CplusplusMode)
	 &&(Parameter->Type->Type == SIR_TYPE_INTERFACE))
	{ ParamBuffer.prepend("(&");	/* interface parameter as reference */
	  ParamBuffer += ")";
	 } /* fi */
     Buffer += Parameter->Type->PrettyString(&ParamBuffer, WriteNotes,
						CplusplusMode);
     Parameter = Parameter->Succ();
     if (Parameter)
	{ Buffer += ", ";
	 } /* fi */
    } /* elihw */
Buffer += ")";
fputs(Type->SubType->PrettyString(&Buffer, WriteNotes, CplusplusMode), File);

if (ferror(File))
   { SIR_ErrMsg.form("Writing to file failed%s",
			GL_SystemErrorMessage());
     return(SIR_ERROR_WRITING_SPECC_FILE);
    } /* fi */

SIR_LineInfo::Add2Tabulator(1);
if ((SIR_Error = FctBody->WriteSC(File, WriteNotes, CplusplusMode)))
   { return(SIR_Error);
    } /* fi */
SIR_LineInfo::Add2Tabulator(-1);

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::WriteSC3 */


ERROR SIR_Symbol::WriteCC2(	/* generates C++ source code */
	FILE		*File,	/* (for class symbols) */
	BOOL		WriteNotes,
	BOOL		AsDeclaration = FALSE)
{
string		Buffer,
		ParamBuffer;
sir_parameter	*Port;
sir_type_ptr	*DeclaredPort;
sir_symbol_ptr	*Interface;
sir_symbol	*ParamSymbol,
		*Instance,
		*Symbol;
sir_portmap	*PortMap;
int		BitBusNum,
		OpenPortNum,
		ConstPortNum;

assert(IsClass());	/* nothing else supported */

if (  ((AsDeclaration) && (ClassScope == NULL))
    ||((! AsDeclaration) && (ClassScope != NULL)))
   { if (LineInfo)
	{ if ((SIR_Error = LineInfo->WriteSC(File, true)))	/* old stuff */
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     else
	{ SIR_LineInfo::WriteNewSC(File, true);	/* new stuff */
	 } /* esle */
    } /* fi */
else
   { SIR_LineInfo::WriteNewSC(File, true);	/* old or new stuff */
    } /* esle */

fputs("class ", File);
fputs(Name.chars(), File);

if (  (! AsDeclaration)
    &&(ClassScope))
   { switch(Class)				/* base classes */
	{ case SIR_SYMBOL_IDENTIFIER:
	  case SIR_SYMBOL_TYPEDEF:
	     { assert(FALSE);	/* this is for classes only */
	      }
	  case SIR_SYMBOL_BEHAVIOR:
	     { if (0 != strcmp(Name.chars(), GL_MAIN_BEHAVIOR_NAME))
		  { fputs(" : public " SIR_SIM_BEHAVIOR_NAME, File);
		   } /* fi */
	       assert(Interfaces != NULL);	/* is a behavior */
	       if (! Interfaces->Empty())
		  { Interface = Interfaces->First();
		    while(Interface)
		       { fputs(", public ", File);
			 fputs(Interface->Symbol->Name.chars(), File);
			 Interface = Interface->Succ();
			} /* elihw */
		   } /* fi */
	       break;
	      }
	  case SIR_SYMBOL_CHANNEL:
	     { fputs(" : public " SIR_SIM_CHANNEL_NAME, File);
	       assert(Interfaces != NULL);	/* is a channel */
	       if (! Interfaces->Empty())
		  { Interface = Interfaces->First();
		    while(Interface)
		       { fputs(", public ", File);
			 fputs(Interface->Symbol->Name.chars(), File);
			 Interface = Interface->Succ();
			} /* elihw */
		   } /* fi */
	       break;
	      }
	  case SIR_SYMBOL_INTERFACE:
	     { /* nothing to inherit from */
	       break;
	      }
	  default:
	     { assert(FALSE);	/* bad symbol class */
	      }
	 } /* hctiws */
     SIR_LineInfo::WriteNL(File);		/* body of the class */
     fputc('{', File);
     SIR_LineInfo::WriteNL(File);
     fputs("private:", File);	/* everything except methods is private */
     SIR_LineInfo::Add2Tabulator(1);
     if (  (Parameters)
	 &&(! Parameters->IsVoid()))
	{ Port = Parameters->First();
	  while(Port)
	     { if (Port->Symbol)
		  { if (Port->Symbol->LineInfo)
		       { if ((SIR_Error = Port->Symbol->LineInfo->WriteSC(
								File, true)))
			    { return(SIR_Error);
			     } /* fi */
			} /* fi */
		    else
		       { SIR_LineInfo::WriteNewSC(File, true);
			} /* esle */
		    Buffer = "(&";
		    Buffer += Port->Symbol->Name.chars();
		    Buffer += ")";
		    fputs(Port->Symbol->Type->PrettyString(&Buffer,
						WriteNotes, TRUE), File);
		    fputc(';', File);
		   } /* fi */
	       Port = Port->Succ();
	      } /* elihw */
	 } /* fi */
     ParamSymbol = ClassScope->First();	/* notes for ports */
     while(ParamSymbol)
	{ if (  (ParamSymbol->IsPort())
	      &&(ParamSymbol->Notes))
	     { if ((SIR_Error = ParamSymbol->Notes->WriteSC(File,
							WriteNotes, TRUE)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  ParamSymbol = ParamSymbol->Succ();
	 } /* elihw */
						/* body */
     /* (1)-(4) User-defined types, declarations and definitions */
     if (ClassScope->UserTypes)
	{ if ((SIR_Error = ClassScope->UserTypes->WriteSC(File,
						WriteNotes, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     /* (5) Type definitions (alias types) */
     Symbol = ClassScope->First();
     while(Symbol)
	{ if (  (Symbol->IsTypeDef())
	      &&((  (Symbol->Type->UserType == NULL)
		  ||(Symbol->Type->UserType->Name != NULL))))
	     { if ((SIR_Error = Symbol->WriteSC(File, WriteNotes, FALSE, TRUE)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  Symbol = Symbol->Succ();
	 } /* elihw */
     /* (6)-(8) NO interface, channel or behavior declarations */
#ifndef NDEBUG
     Symbol = ClassScope->First();
     while(Symbol)
	{ assert((  (Symbol->IsInterface())
		  ||(Symbol->IsChannel())
		  ||(Symbol->IsBehavior())) == FALSE);
	  Symbol = Symbol->Succ();
	 } /* elihw */
#endif /* NDEBUG */

     SIR_LineInfo::Add2Tabulator(-1);
     SIR_LineInfo::WriteNL(File);
     fputs("public:", File);	/* everything except methods is private */
     SIR_LineInfo::Add2Tabulator(1);
     if (! IsInterface())
	{ /* (9a) Declaration of constructor */
	  SIR_LineInfo::WriteNewSC(File, true);
	  fputs(Name.chars(), File);
	  if (  (Parameters)		/* ports */
	      &&(! Parameters->IsVoid()))
	     { fputc('(', File);
	       Port = Parameters->First();
	       while(Port)
		  { if (Port->Symbol)
		       { Buffer = "(&";	/* as reference */
			 Buffer += Port->Symbol->Name;
			 Buffer += ")";
			} /* fi */
		    else
		       { Buffer = "&";
			} /* esle */
		    fputs(Port->Type->PrettyString(&Buffer, WriteNotes, TRUE),
						File);
		    Port = Port->Succ();
		    if (Port)
		       { fputs(", ", File);
			} /* fi */
		   } /* elihw */
	       fputs(");", File);
	      } /* fi */
	  else
	     { fputs("(void);", File);
	      } /* esle */
	  /* (9b) Declaration of destructor */
	  SIR_LineInfo::WriteNewSC(File, true);
	  fputs("virtual ~", File);
	  fputs(Name.chars(), File);
	  fputs("(void);", File);
	 } /* fi */
     /* (9c) Declarations of (variables and) other methods */
     Symbol = ClassScope->First();
     while(Symbol)
	{ if (  (Symbol->IsVariable())
	      ||(Symbol->IsFunction()))
	     { if ((SIR_Error = Symbol->WriteSC(File, WriteNotes, TRUE, TRUE)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  Symbol = Symbol->Succ();
	 } /* elihw */
     SIR_LineInfo::Add2Tabulator(-1);
     SIR_LineInfo::WriteNL(File);
     fputs("private:", File);	/* everything except methods is private */
     SIR_LineInfo::Add2Tabulator(1);
     /* (10) Variable definitions and class instantiations */
     Symbol = ClassScope->First();		/* variable definitions */
     while(Symbol)
	{ if (Symbol->IsVariableDefinition())
	     { if ((SIR_Error = Symbol->WriteSC(File, WriteNotes, FALSE, TRUE)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  Symbol = Symbol->Succ();
	 } /* elihw */
     Instance = ClassScope->First();		/* sliced bus declarations */
     BitBusNum = 0;			/* (must appear after variable defs. */
     OpenPortNum = 0;			/* and before class instantiations!) */
     ConstPortNum = 0;
     while(Instance)
	{ if (  (Instance->IsBehaviorInstance())
	      ||(Instance->IsChannelInstance()))
	     { assert(Instance->PortMappings != NULL);	/* must exist */
	       PortMap = Instance->PortMappings->First();
	       DeclaredPort = Instance->Type->Parameters->First();
	       while(PortMap)
		  { assert(DeclaredPort != NULL);	/* must exist */
		    PortMap->WriteCC_Decl(File, DeclaredPort->Type,
				&BitBusNum, &OpenPortNum, &ConstPortNum);
		    PortMap = PortMap->Succ();
		    DeclaredPort = DeclaredPort->Succ();
		   } /* elihw */
	      } /* fi */
	  Instance = Instance->Succ();
	 } /* elihw */
     Symbol = ClassScope->First();		/* class instantiations */
     while(Symbol)
	{ if (  (Symbol->IsBehaviorInstance())
	      ||(Symbol->IsChannelInstance()))
	     { if ((SIR_Error = Symbol->WriteSC(File, WriteNotes, FALSE, TRUE)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  Symbol = Symbol->Succ();
	 } /* elihw */
     SIR_LineInfo::Add2Tabulator(-1);
     SIR_LineInfo::WriteNL(File);
     fputc('}', File);
    } /* fi */

fputc(';', File);

if (ferror(File))
   { SIR_ErrMsg.form("Writing to file failed%s",
			GL_SystemErrorMessage());
     return(SIR_ERROR_WRITING_SPECC_FILE);
    } /* fi */

if (AsDeclaration)
   { if (Notes)
	{ if ((SIR_Error = Notes->WriteSC(File, WriteNotes, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::WriteCC2 */


ERROR SIR_Symbol::WriteCC2b(	/* generates C++ source code */
	FILE		*File,	/* (for classes, method implementations) */
	BOOL		WriteNotes)
{
string		Buffer,
		ParamBuffer;
sir_parameter	*Port;
sir_type_ptr	*DeclaredPort;
sir_parameter	*Parameter;
sir_symbol	*Instance,
		*MainMethod,
		*Symbol;
sir_portmap	*PortMap;
const char	*TmpString;
BOOL		PutColon,
		PutNewLine;
int		ParamNum;
int		BitBusNum,
		OpenPortNum,
		ConstPortNum;

assert(IsClass());	/* nothing else supported */
assert(ClassScope != NULL);

     /* (a) constructor implementation */

     SIR_LineInfo::WriteNewSC(File, true);	/* constructor */
     fputs(Name.chars(), File);
     fputs("::", File);
     fputs(Name.chars(), File);
     if (  (Parameters)			/* ports */
	 &&(! Parameters->IsVoid()))
	{ fputc('(', File);
	  Port = Parameters->First();
	  while(Port)
	     { if (Port->Symbol)
		  { Buffer = "(&";	/* as reference */
		    Buffer += Port->Symbol->Name;
		    Buffer += ")";
		   } /* fi */
	       else
		  { Buffer = "&";
		   } /* esle */
	       fputs(Port->Type->PrettyString(&Buffer, WriteNotes, TRUE), File);
	       Port = Port->Succ();
	       if (Port)
		  { fputs(", ", File);
		   } /* fi */
	      } /* elihw */
	  fputc(')', File);
	 } /* fi */
     else
	{ fputs("(void)", File);
	 } /* esle */
     SIR_LineInfo::Add2Tabulator(1);
     PutColon = TRUE;
     PutNewLine = TRUE;
     if (Parameters)			/* port mappings */
	{ Port = Parameters->First();
	  while(Port)
	     { if (Port->Symbol)
		  { if (PutNewLine)
		       { SIR_LineInfo::WriteNL(File);
			 PutNewLine = FALSE;
			} /* fi */
		    if (PutColon)
		       { fputs(": ", File);
			 PutColon = FALSE;
			} /* fi */
		    else
		       { fputs(", ", File);
			} /* esle */
		    fputs(Port->Symbol->Name.chars(), File);
		    fputc('(', File);
		    fputs(Port->Symbol->Name.chars(), File);
		    fputc(')', File);
		   } /* fi */
	       Port = Port->Succ();
	      } /* elihw */
	 } /* fi */
     Symbol = ClassScope->First();	/* member initializer */
     while(Symbol)
	{ if (  (Symbol->IsVariableDefinition())
	      &&(  (Symbol->Initializer)
		 ||(  (Symbol->StorageClass == SIR_STORAGE_PIPED)
		    &&(Symbol->PipeStages > 1))))
	     { if (PutColon)
		  { SIR_LineInfo::WriteNL(File);
		    fputs(": ", File);
		    PutColon = FALSE;
		   } /* fi */
	       else
		  { fputc(',', File);
		    SIR_LineInfo::WriteNL(File);
		   } /* esle */
	       fputs(Symbol->Name.chars(), File);
	       fputc('(', File);
	       if (Symbol->Initializer)
		  { assert(Symbol->StorageClass != SIR_STORAGE_PIPED);
		    assert(Symbol->Initializer->InitList
						== NULL); /* unsupported */
		    fputs(Symbol->Initializer->Initializer->Print(TRUE), File);
		   } /* fi */
	       else	/* Symbol->StorageClass == SIR_STORAGE_PIPED */
		  { fprintf(File, "%d", Symbol->PipeStages);
		   } /* esle */
	       fputc(')', File);
	      } /* fi */
	  Symbol = Symbol->Succ();
	 } /* elihw */
     Instance = ClassScope->First();	/* sliced bus initializations */
     BitBusNum = 0;
     OpenPortNum = 0;
     ConstPortNum = 0;
     while(Instance)
	{ if (  (Instance->IsBehaviorInstance())
	      ||(Instance->IsChannelInstance()))
	     { assert(Instance->PortMappings != NULL);	/* must exist */
	       PortMap = Instance->PortMappings->First();
	       DeclaredPort = Instance->Type->Parameters->First();
	       while(PortMap)
		  { assert(DeclaredPort != NULL);
		    if ((TmpString = PortMap->WriteCC_Init(DeclaredPort->Type,
				&BitBusNum, &OpenPortNum, &ConstPortNum)))
		       { if (PutColon)
			    { SIR_LineInfo::WriteNL(File);
			      fputs(": ", File);
			      PutColon = FALSE;
			     } /* fi */
			 else
			    { fputc(',', File);
			      SIR_LineInfo::WriteNL(File);
			     } /* esle */
			 fputs(TmpString, File);
			} /* fi */
		    PortMap = PortMap->Succ();
		    DeclaredPort = DeclaredPort->Succ();
		   } /* elihw */
	      } /* fi */
	  Instance = Instance->Succ();
	 } /* elihw */
     Instance = ClassScope->First();	/* instance port mappings */
     BitBusNum = 0;
     OpenPortNum = 0;
     ConstPortNum = 0;
     while(Instance)
	{ if (  (Instance->IsBehaviorInstance())
	      ||(Instance->IsChannelInstance()))
	     { assert(Instance->PortMappings != NULL);	/* must exist */
	       if (! Instance->PortMappings->Empty())	/* but can be empty */
		  { if (PutColon)
		       { SIR_LineInfo::WriteNL(File);
			 fputs(": ", File);
			 PutColon = FALSE;
			} /* fi */
		    else
		       { fputc(',', File);
			 SIR_LineInfo::WriteNL(File);
			} /* esle */
		    fputs(Instance->Name.chars(), File);
		    fputc('(', File);
		    PortMap = Instance->PortMappings->First();
		    DeclaredPort = Instance->Type->Parameters->First();
		    while(PortMap)
		       { assert(DeclaredPort != NULL);
			 PortMap->WriteCC(File, DeclaredPort->Type,
				&BitBusNum, &OpenPortNum, &ConstPortNum);
			 PortMap = PortMap->Succ();
			 DeclaredPort = DeclaredPort->Succ();
			 if (PortMap)
			    { fputs(", ", File);
			     } /* fi */
			} /* elihw */
		    fputc(')', File);
		   } /* fi */
	      } /* fi */
	  Instance = Instance->Succ();
	 } /* elihw */
     SIR_LineInfo::WriteNewSC(File, true);
     fputs("{ /* nothing */ }", File);
     SIR_LineInfo::Add2Tabulator(-1);

     /* (b) destructor implementation */

     SIR_LineInfo::WriteNewSC(File, true);	/* destructor */
     fputs(Name.chars(), File);
     fputs("::~", File);
     fputs(Name.chars(), File);
     fputs("(void)", File);
     SIR_LineInfo::Add2Tabulator(1);
     SIR_LineInfo::WriteNL(File);
     fputs("{ /* nothing */ }", File);
     SIR_LineInfo::Add2Tabulator(-1);

     /* (c) other methods implementations */

     Symbol = ClassScope->First();
     while(Symbol)
	{ if (Symbol->IsFunctionDefinition())
	     { if ((SIR_Error = Symbol->WriteSC3(File, WriteNotes, TRUE)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  Symbol = Symbol->Succ();
	 } /* elihw */


if (  (IsBehaviorDefinition())			/* main function */
    &&(0 == strcmp(Name.chars(), GL_MAIN_BEHAVIOR_NAME)))
   { assert(GetScope()->ScopeInfo == SIR_SCOPE_GLOBAL);
#ifdef SIR_STORE_MAIN_BEHAVIOR_AS_STATIC
     SIR_LineInfo::WriteNewSC(File, true);
     fputs(GL_MAIN_BEHAVIOR_NAME " " SIR_CXX_MAIN_INST_NAME ";", File);
     SIR_LineInfo::WriteNL(File);
#endif /* SIR_STORE_MAIN_BEHAVIOR_AS_STATIC */
     if (GetScope()->FindLocally(GL_MAIN_FUNCTION_NAME))
	{ GL_PrintWarning(GL_WARN_IMPORTANT,
		"Skipping generation of function '" GL_MAIN_FUNCTION_NAME "';"
							GL_WARN_MSG_NEWLINE
		"global symbol '" GL_MAIN_FUNCTION_NAME "' already exists");
	 } /* fi */
     else
	{ MainMethod = ClassScope->FindLocally(GL_MAIN_METHOD_NAME);
	  assert(  (MainMethod != NULL)	/* semantic checks assure this */
		 &&(MainMethod->IsFunctionDefinition()));
	  SIR_LineInfo::WriteNewSC(File, true);
	  Buffer = GL_MAIN_FUNCTION_NAME "(";
	  Parameter = MainMethod->Parameters->First();
	  ParamNum = 0;
	  while(Parameter)
	     { if (Parameter->Symbol)
		  { Buffer += Parameter->Symbol->PrintStorageClass(TRUE);
		    ParamBuffer = Parameter->Symbol->Name;
		   } /* fi */
	       else
		  { if (Parameter->Type->Type == SIR_TYPE_VOID)
		       { ParamBuffer = "";
			} /* fi */
		    else
		       { ParamBuffer.form(SIR_CXX_MAIN_ARG_NAME, ParamNum);
			} /* esle */
		   } /* esle */
	       Buffer += Parameter->Type->PrettyString(&ParamBuffer,
							FALSE, TRUE);
	       Parameter = Parameter->Succ();
	       ParamNum++;
	       if (Parameter)
		  { Buffer += ", ";
		   } /* fi */
	      } /* elihw */
	  Buffer += ")";
	  fputs(MainMethod->Type->SubType->Modified(false,
				MainMethod->Type->Volatile)->PrettyString(
							&Buffer, FALSE, TRUE),
			File);
	  SIR_LineInfo::Add2Tabulator(1);
	  SIR_LineInfo::WriteNL(File);
	  fputc('{', File);
#ifdef SIR_STORE_MAIN_BEHAVIOR_ON_STACK
	  SIR_LineInfo::WriteNL(File);
	  fputs(GL_MAIN_BEHAVIOR_NAME " " SIR_CXX_MAIN_INST_NAME ";", File);
#endif /* SIR_STORE_MAIN_BEHAVIOR_ON_STACK */
	  SIR_LineInfo::WriteNL(File);
	  if (MainMethod->Type->SubType->Type != SIR_TYPE_VOID)
	     { Buffer = SIR_CXX_MAIN_RET_NAME;
	       fputs(MainMethod->Type->SubType->Modified(false,
				MainMethod->Type->Volatile)->PrettyString(
							&Buffer, FALSE, TRUE),
			File);
	       fputc(';', File);
	       SIR_LineInfo::WriteNL(File);
	      } /* fi */
	  SIR_LineInfo::WriteNL(File);
	  fputs(SIR_SIM_INIT "();", File);
	  SIR_LineInfo::WriteNL(File);
	  if (MainMethod->Type->SubType->Type != SIR_TYPE_VOID)
	     { fputs(SIR_CXX_MAIN_RET_NAME " = ", File);
	      } /* fi */
	  fputs(SIR_CXX_MAIN_INST_NAME "." GL_MAIN_METHOD_NAME "(", File);
	  Parameter = MainMethod->Parameters->First();
	  ParamNum = 0;
	  while(Parameter)
	     { if (Parameter->Symbol)
		  { fputs(Parameter->Symbol->Name.chars(), File);
		   } /* fi */
	       else
		  { if (Parameter->Type->Type != SIR_TYPE_VOID)
		       { fprintf(File, SIR_CXX_MAIN_ARG_NAME, ParamNum);
			} /* fi */
		   } /* esle */
	       Parameter = Parameter->Succ();
	       ParamNum++;
	       if (Parameter)
		  { fputs(", ", File);
		   } /* fi */
	      } /* elihw */
	  fputs(");", File);
	  SIR_LineInfo::WriteNL(File);
	  fputs(SIR_SIM_EXIT "();", File);
	  SIR_LineInfo::WriteNL(File);
	  if (MainMethod->Type->SubType->Type != SIR_TYPE_VOID)
	     { fputs("return(" SIR_CXX_MAIN_RET_NAME ");", File);
	       SIR_LineInfo::WriteNL(File);
	      } /* fi */
	  fputc('}', File);
	  SIR_LineInfo::WriteNL(File);
	  SIR_LineInfo::Add2Tabulator(-1);
	 } /* esle */
    } /* fi */

if (ferror(File))
   { SIR_ErrMsg.form("Writing to file failed%s",
			GL_SystemErrorMessage());
     return(SIR_ERROR_WRITING_SPECC_FILE);
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::WriteCC2b */


ERROR SIR_Symbol::WriteCC2IP(	/* generates C++ source code */
	FILE		*File)	/* (for IP classes) */
{
string		Buffer,
		ParamBuffer;
sir_type_ptr	*PortTypePtr;
sir_symbol_ptr	*Interface;
sir_symbol	*Symbol,
		*Method;
int		PortNum;

assert(IsBehavior() || IsChannel());	/* nothing else supported */
assert(ClassScope == NULL);	/* otherwise it's not an IP */
assert(Interfaces != NULL);	/* is a behavior or channel */

Symbol = GetScope()->First();
while(Symbol)
   { if (  (  (Symbol->IsBehavior())
	    ||(Symbol->IsChannel()))
	 &&(Symbol->Instantiates(this)))
	{ break;	/* there's at least one instantiation */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */
if (! Symbol)		/* no instance found */
   { return(SIR_ERROR_NO_ERROR);	/* unused */
    } /* fi */

if (LineInfo)
   { if ((SIR_Error = LineInfo->WriteSC(File, true)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
else
   { SIR_LineInfo::WriteNewSC(File, true);
    } /* esle */

fputs("class ", File);
fputs(Name.chars(), File);

if (Class == SIR_SYMBOL_BEHAVIOR)	/* base classes */
   { if (0 != strcmp(Name.chars(), GL_MAIN_BEHAVIOR_NAME))
	{ fputs(" : public " SIR_SIM_BEHAVIOR_NAME, File);
	 } /* fi */
    } /* fi */
else
   { fputs(" : public " SIR_SIM_CHANNEL_NAME, File);
    } /* esle */
if (! Interfaces->Empty())
   { Interface = Interfaces->First();
     while(Interface)
	{ fputs(", public ", File);
	  fputs(Interface->Symbol->Name.chars(), File);
	  Interface = Interface->Succ();
	 } /* elihw */
    } /* fi */

SIR_LineInfo::WriteNL(File);		/* _fake_ body of the class */
fputc('{', File);
SIR_LineInfo::WriteNL(File);
fputs("private:", File);	/* everything except methods is private */
SIR_LineInfo::Add2Tabulator(1);
if (! Type->VoidParameters())		/* dummy ports */
   { PortNum = 0;
     PortTypePtr = Type->Parameters->First();
     while(PortTypePtr)
	{ SIR_LineInfo::WriteNewSC(File, true);
	  Buffer.form("(&" SIR_CXX_DUMMY_PORT_NAME ")", PortNum++);
	  fputs(PortTypePtr->Type->PrettyString(&Buffer, FALSE, TRUE), File);
	  fputc(';', File);
	  PortTypePtr = PortTypePtr->Succ();
	 } /* elihw */
    } /* fi */

SIR_LineInfo::Add2Tabulator(-1);
SIR_LineInfo::WriteNL(File);
fputs("public:", File);		/* everything except methods is private */
SIR_LineInfo::Add2Tabulator(1);
SIR_LineInfo::WriteNewSC(File, true);	/* constructor declaration */
fputs(Name.chars(), File);
if (! Type->VoidParameters())		/* dummy ports */
   { fputc('(', File);
     PortTypePtr = Type->Parameters->First();
     while(PortTypePtr)
	{ Buffer = "&";	/* as reference */
	  fputs(PortTypePtr->Type->PrettyString(&Buffer, FALSE, TRUE), File);
	  PortTypePtr = PortTypePtr->Succ();
	  if (PortTypePtr)
	     { fputs(", ", File);
	      } /* fi */
	 } /* elihw */
     fputs(");", File);
    } /* fi */
else
   { fputs("(void);", File);
    } /* esle */

SIR_LineInfo::WriteNewSC(File, true);	/* destructor declaration */
fputs("virtual ~", File);
fputs(Name.chars(), File);
fputs("(void);", File);

#ifndef NDEBUG
Interface = Interfaces->First();	/* implemented methods declarations */
while(Interface)
   { if (Interface->Symbol->ClassScope)
	{ Method = Interface->Symbol->ClassScope->First();
	  while(Method)
	     { assert(Method->Color == SIR_WHITE);	/* unmarked */
	       Method = Method->Succ();
	      } /* elihw */
	 } /* fi */
     Interface = Interface->Succ();
    } /* elihw */
#endif /* NDEBUG */
if (IsBehavior())
   { SIR_LineInfo::WriteNewSC(File, true);
     fputs("void " GL_MAIN_METHOD_NAME "(void);", File);
     MarkImplementedMethods(GL_MAIN_METHOD_NAME, Interfaces->First());
    } /* fi */
Interface = Interfaces->First();
while(Interface)
   { if (Interface->Symbol->ClassScope)
	{ Method = Interface->Symbol->ClassScope->First();
	  while(Method)
	     { if (Method->Color == SIR_WHITE)	/* not already declared */
		  { assert(Method->IsFunction());
		    SIR_LineInfo::WriteNewSC(File, true);
		    Buffer = Method->Name;
		    fputs(Method->Type->PrettyString(&Buffer, FALSE, TRUE),
				File);
		    fputc(';', File);
		    Method->Color = SIR_RED;
		    MarkImplementedMethods(Method->Name.chars(),
						Interface->Succ());
		   } /* fi */
	       Method = Method->Succ();
	      } /* elihw */
	 } /* fi */
     Interface = Interface->Succ();
    } /* elihw */
Interface = Interfaces->First();	/* clean up */
while(Interface)
   { if (Interface->Symbol->ClassScope)
	{ Method = Interface->Symbol->ClassScope->First();
	  while(Method)
	     { assert(Method->Color == SIR_RED);/* must have been declared */
	       Method->Color = SIR_WHITE;	/* remove markers */
	       Method = Method->Succ();
	      } /* elihw */
	 } /* fi */
     Interface = Interface->Succ();
    } /* elihw */

SIR_LineInfo::Add2Tabulator(-1);
SIR_LineInfo::WriteNL(File);
fputs("};", File);

if (ferror(File))
   { SIR_ErrMsg.form("Writing to file failed%s",
			GL_SystemErrorMessage());
     return(SIR_ERROR_WRITING_SPECC_FILE);
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::WriteCC2IP */


ERROR SIR_Symbol::WriteSC4ClassBody(	/* generates SpecC for class bodies */
	FILE		*File,		/* (for class definitions only) */
	BOOL		WriteNotes = TRUE,
	BOOL		WriteLines = TRUE,
	const char	*ThisFilename = NULL,	/* default location if */
	unsigned int	ThisLineNumber = 1)	/* line info unknown   */
{
sir_symbol	*ParamSymbol;

assert(File != NULL);
assert(ClassScope != NULL);	/* must be a class definition */

SIR_LineInfo::InitWriteSC(WriteLines,	/* init for new file */
			ThisFilename,
			ThisLineNumber);

if (ParamScope)
   { ParamSymbol = ParamScope->First();
     while(ParamSymbol)
        { if (ParamSymbol->Notes)
	     { if ((SIR_Error = ParamSymbol->Notes->WriteSC(File, WriteNotes)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  ParamSymbol = ParamSymbol->Succ();
	 } /* elihw */
    } /* fi */
if ((SIR_Error = ClassScope->WriteSC(File, WriteNotes)))
   { return(SIR_Error);
    } /* fi */
SIR_LineInfo::WriteNL(File);

if (ferror(File))
   { SIR_ErrMsg.form("Writing to file failed%s",
			GL_SystemErrorMessage());
     return(SIR_ERROR_WRITING_SPECC_FILE);
    } /* fi */

return(SIR_ERROR_NO_ERROR);	/* successful */

} /* end of SIR_Symbol::WriteSC4ClassBody */


ERROR SIR_Symbol::CheckInitializer(	/* checks whether initializer is legal*/
	sir_initializer	*Initializer = NULL,	/* (if not specified, */
	sir_type	*Type = NULL,		/*  own data is used) */
	SIR_STORAGE	*StorageClass = NULL,
	int		*PipeStages = NULL)
{
sir_initializer	*CheckInit;
sir_type	*CheckType;
SIR_STORAGE	CheckStorage;
int		CheckPipes;

if (Initializer)
   { CheckInit = Initializer;
    } /* fi */
else
   { CheckInit = this->Initializer;
    } /* esle */
if (Type)
   { CheckType = Type;
    } /* fi */
else
   { CheckType = this->Type;
    } /* esle */
if (StorageClass)
   { CheckStorage = *StorageClass;
    } /* fi */
else
   { CheckStorage = this->StorageClass;
    } /* esle */
if (PipeStages)
   { CheckPipes = *PipeStages;
    } /* fi */
else
   { CheckPipes = this->PipeStages;
    } /* esle */

if (! Initializer)
   { return(SIR_ERROR_NO_ERROR);	/* empty is always ok */
    } /* fi */

if (  (CheckStorage == SIR_STORAGE_TYPEDEF)
    ||(CheckStorage == SIR_STORAGE_EXTERN)
    ||(CheckStorage == SIR_STORAGE_PIPED))
   { SIR_ErrMsg.form("Storage class '%s' forbids initialization"
							GL_ERROR_MSG_NEWLINE
			"for variable '%s'",
			SIR_Symbol::PrintStorageClass(CheckStorage),
			Name.chars());
     return(SIR_ERROR_FORBIDDEN_INITIALIZER);
    } /* fi */

if (  (CheckInit->InitList)
    &&(GetScope()->ScopeInfo == SIR_SCOPE_CLASS))
   { SIR_ErrMsg.form("Complex initializer not supported" GL_ERROR_MSG_NEWLINE
			"for class member '%s'",
			Name.chars());
     return(SIR_ERROR_COMPLEX_MEMBER_INITIALIZER);
    } /* fi */

if ((SIR_Error = CheckType->CheckInitializer(CheckInit)))
   { return(SIR_Error);
    } /* fi */

return(SIR_ERROR_NO_ERROR);	/* is acceptable */

} /* end of SIR_Symbol::CheckInitializer */


ERROR SIR_Symbol::CheckStorageClass(	/* checks whether storage class legal */
	SIR_STORAGE	*StorageClass = NULL,	/* (if not specified, */
	int		*PipeStages = NULL,	/*  own data is used) */
	SIR_SYMBOLCLASS	*Class = NULL,
	SIR_SCOPE	*ScopeInfo = NULL)
{
SIR_STORAGE	CheckStorage;
int		CheckPipes;
SIR_SYMBOLCLASS	CheckClass;
SIR_SCOPE	CheckScope;

if (StorageClass)
   { CheckStorage = *StorageClass;
    } /* fi */
else
   { CheckStorage = this->StorageClass;
    } /* esle */
if (PipeStages)
   { CheckPipes = *PipeStages;
    } /* fi */
else
   { CheckPipes = this->PipeStages;
    } /* esle */
if (Class)
   { CheckClass = *Class;
    } /* fi */
else
   { CheckClass = this->Class;
    } /* esle */
if (ScopeInfo)
   { CheckScope = *ScopeInfo;
    } /* fi */
else
   { CheckScope = GetScope()->ScopeInfo;
    } /* esle */

switch(CheckClass)
   { case SIR_SYMBOL_IDENTIFIER:
	{ switch(CheckScope)
	     { case SIR_SCOPE_GLOBAL:
		  { if (  (CheckStorage == SIR_STORAGE_AUTO)
			||(CheckStorage == SIR_STORAGE_REGISTER)
			||(CheckStorage == SIR_STORAGE_PIPED))
		       { SIR_ErrMsg.form(
			"Storage class '%s' not allowed in global scope",
				PrintStorageClass(CheckStorage, CheckPipes));
			 return(SIR_ERROR_ILLEGAL_STORAGE_CLASS);
			} /* fi */
		    if (  (IsFunction())
			&&(CheckStorage == SIR_STORAGE_TYPEDEF))
		       { SIR_ErrMsg.form(
			"Storage class '%s' not allowed for functions",
				PrintStorageClass(CheckStorage, CheckPipes));
			 return(SIR_ERROR_ILLEGAL_STORAGE_CLASS);
			} /* fi */
		    break;
		   }
	       case SIR_SCOPE_CLASS:
		  { if (  (IsPort())
			&&(CheckStorage != SIR_STORAGE_NONE))
		       { SIR_ErrMsg.form(
			"Storage class '%s' not allowed for ports",
				PrintStorageClass(CheckStorage, CheckPipes));
			 return(SIR_ERROR_ILLEGAL_STORAGE_CLASS);
			} /* fi */
		    if (  (CheckStorage == SIR_STORAGE_STATIC)
			||(CheckStorage == SIR_STORAGE_EXTERN)
			||(CheckStorage == SIR_STORAGE_AUTO)
			||(CheckStorage == SIR_STORAGE_REGISTER))
		       { SIR_ErrMsg.form(
			"Storage class '%s' not allowed in class scope",
				PrintStorageClass(CheckStorage, CheckPipes));
			 return(SIR_ERROR_ILLEGAL_STORAGE_CLASS);
			} /* fi */
		    if (  (IsFunction())
			&&(  (CheckStorage == SIR_STORAGE_PIPED)
			   ||(CheckStorage == SIR_STORAGE_TYPEDEF)))
		       { SIR_ErrMsg.form(
			"Storage class '%s' not allowed for functions",
				PrintStorageClass(CheckStorage, CheckPipes));
			 return(SIR_ERROR_ILLEGAL_STORAGE_CLASS);
			} /* fi */
		    break;
		   }
	       case SIR_SCOPE_PARAMETER:
		  { if (  (CheckStorage == SIR_STORAGE_TYPEDEF)
			||(CheckStorage == SIR_STORAGE_EXTERN)
			||(CheckStorage == SIR_STORAGE_STATIC)
			||(CheckStorage == SIR_STORAGE_PIPED))
		       { SIR_ErrMsg.form(
			"Storage class '%s' not allowed for parameters",
				PrintStorageClass(CheckStorage, CheckPipes));
			 return(SIR_ERROR_ILLEGAL_STORAGE_CLASS);
			} /* fi */
		    break;
		   }
	       case SIR_SCOPE_STATEMENT:
		  { break;
		   }
	       case SIR_SCOPE_USERTYPE:
		  { assert(CheckStorage == SIR_STORAGE_NONE);
		    break;
		   }
	       default:
		  { assert(FALSE);	/* illegal scope info */
		   }
	      } /* hctiws */
	  break;
	 }
     case SIR_SYMBOL_TYPEDEF:
	{ assert(CheckStorage == SIR_STORAGE_TYPEDEF);
	  break;
	 }
     case SIR_SYMBOL_BEHAVIOR:
	{ assert(CheckStorage == SIR_STORAGE_NONE);
	  break;
	 }
     case SIR_SYMBOL_CHANNEL:
	{ assert(CheckStorage == SIR_STORAGE_NONE);
	  break;
	 }
     case SIR_SYMBOL_INTERFACE:
	{ assert(CheckStorage == SIR_STORAGE_NONE);
	  break;
	 }
     default:
	{ assert(FALSE);	/* bad symbol class */
	 }
    } /* hctiws */

return(SIR_ERROR_NO_ERROR);	/* looks ok */

} /* end of SIR_Symbol::CheckStorageClass */


ERROR SIR_Symbol::Annotate(	/* attach a note to the symbol */
	sir_note	*NewNote)	/* (consumes NewNote) */
{

assert(NewNote != NULL);

NewNote->Symbol = this;		/* symbol note */
NewNote->UserType = NULL;
NewNote->Label = NULL;

if (Notes)
   { if (! Notes->Insert(NewNote))
	{ delete NewNote;
	  return(SIR_Error);
	 } /* fi */
    } /* fi */
else
   { Notes = new SIR_Notes(NewNote);
    } /* esle */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbol::Annotate */


void SIR_Symbol::Touch(		/* modification update (remove import info) */
	BOOL		ImportOnly = FALSE)/* (dflt: remove source info too) */
{

if (Imported)		/* remove import information */
   { Imported->TakeOverAndRemove(GetDesign());
    } /* fi */

if (! ImportOnly)	/* remove source code info */
   { SIR_Node::Strip();
    } /* fi */

} /* end of SIR_Symbol::Touch */


void SIR_Symbol::Strip(		/* remove source location for the tree below */
	BOOL		LocalOnly = FALSE) /* (dflt: work recursively) */
{

if (LocalOnly)
   { SIR_Node::Strip();
    } /* fi */
else
   { DFS_ForAllNodes(&SIR_Node::Strip, NULL);
    } /* esle */

} /* end of SIR_Symbol::Strip */


	/*******************/
	/*** SIR_Symbols ***/
	/*******************/


//++++++++++++++++++++++++++++ API Layer 1 +++++++++++++++++++++++++++++//


SIR_Symbols::SIR_Symbols(		/* constructor #1 */
	sir_symbols	*Parent,	/* (creates UserTypes automatically) */
	SIR_SCOPE	ScopeInfo,
	sir_symbol	*ParentSymbol = NULL,
	sir_usertype	*ParentUType = NULL,
	sir_statement	*ParentStmnt = NULL)
{

if (Parent)
   { SIR_Symbols::Parent	= Parent;
     SIR_Symbols::UserTypes	= new SIR_UserTypes(Parent->UserTypes, this);
    } /* fi */
else
   { SIR_Symbols::Parent	= NULL;
     SIR_Symbols::UserTypes	= new SIR_UserTypes(NULL, this);
    } /* esle */
SIR_Symbols::ScopeInfo		= ScopeInfo;
SIR_Symbols::ParentSymbol	= ParentSymbol;
SIR_Symbols::ParentUType	= ParentUType;
SIR_Symbols::ParentStmnt	= ParentStmnt;

} /* end of SIR_Symbols::SIR_Symbols #1 */


SIR_Symbols::SIR_Symbols(		/* constructor #3 (duplicator) */
	sir_symbols	*Original)
{
sir_symbol	*Curr;

Curr = Original->First();
while(Curr)
   { Append(new SIR_Symbol(Curr));
     Curr = Curr->Succ();
    } /* elihw */

SIR_Symbols::Parent		= Original->Parent;
SIR_Symbols::UserTypes		= Original->UserTypes ?
					new SIR_UserTypes(
						Original->UserTypes) :
					NULL;
SIR_Symbols::ScopeInfo		= Original->ScopeInfo;
SIR_Symbols::ParentSymbol	= Original->ParentSymbol;
SIR_Symbols::ParentUType	= Original->ParentUType;
SIR_Symbols::ParentStmnt	= Original->ParentStmnt;

} /* end of SIR_Symbols::SIR_Symbols #3 */


SIR_Symbols::~SIR_Symbols(void)		/* destructor */
{

delete UserTypes;

} /* end of SIR_Symbols::~SIR_Symbols */


BOOL SIR_Symbols::IsAncestorOf(	/* checks if the symbol can see this scope */
	sir_symbol	*Symbol)
{
sir_symbols	*Scope;

assert(Symbol != NULL);

Scope = Symbol->GetScope();
while(Scope)
   { if (Scope == this)
	{ return(TRUE);
	 } /* fi */
     Scope = Scope->Parent;
    } /* elihw */

return(FALSE);

} /* end of SIR_Symbols::IsAncestorOf */


ERROR SIR_Symbols::DFS_ForAllNodes(	/* iterator for over all nodes */
	sir_node_mptr	MemberFct,	/* (depth first) */
	sir_node_marg	MemberFctArg)
{
sir_symbol	*Symbol,
		*Succ;

/* this is not a node, but there are nodes below */

if (UserTypes)
   { if ((SIR_Error = UserTypes->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */

Symbol = First();
while(Symbol)
   { Succ = Symbol->Succ();
     if ((SIR_Error = Symbol->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
     Symbol = Succ;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::DFS_ForAllNodes */


ERROR SIR_Symbols::DFS_ForAllSymbols(	/* iterator over all symbols */
	sir_symbol_mptr	MemberFct,	/* (depth first) */
	sir_symbol_marg	MemberFctArg)
{
sir_symbol	*Symbol,
		*Succ;

/* this is not a symbol, but there are symbols below */

if (UserTypes)
   { if ((SIR_Error = UserTypes->DFS_ForAllSymbols(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */

Symbol = First();
while(Symbol)
   { Succ = Symbol->Succ();
     if ((SIR_Error = Symbol->DFS_ForAllSymbols(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
     Symbol = Succ;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::DFS_ForAllSymbols */


ERROR SIR_Symbols::DFS_ForAllUserTypes(	/* iterator over all usertypes */
	sir_usertp_mptr	MemberFct,	/* (depth first) */
	sir_usertp_marg	MemberFctArg)
{
sir_symbol	*Symbol,
		*Succ;

/* this is not a usertype, but there are usertypes below */

if (UserTypes)
   { if ((SIR_Error = UserTypes->DFS_ForAllUserTypes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */

Symbol = First();
while(Symbol)
   { Succ = Symbol->Succ();
     if ((SIR_Error = Symbol->DFS_ForAllUserTypes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
     Symbol = Succ;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::DFS_ForAllUserTypes */


ERROR SIR_Symbols::DFS_ForAllNotes(	/* iterator over all notes */
	sir_note_mptr	MemberFct,	/* (depth first) */
	sir_note_marg	MemberFctArg)
{
sir_symbol	*Symbol,
		*Succ;

/* this is not a note, but there are notes below */

if (UserTypes)
   { if ((SIR_Error = UserTypes->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */

Symbol = First();
while(Symbol)
   { Succ = Symbol->Succ();
     if ((SIR_Error = Symbol->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
     Symbol = Succ;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::DFS_ForAllNotes */


BOOL SIR_Symbols::DFS_FindDependant(	/* searches for dependants (DFS) */
	sir_symbol	**SkipSymbol,
	sir_symbol	*ThatSymbol,
	sir_symbol	**DepSymbol,
	sir_statement	**DepStmnt,
	sir_expression	**DepExpr,
	SIR_DEPENDENCY	*Reason)
{
sir_symbol	*Symbol;

assert(SkipSymbol != NULL);
assert(ThatSymbol != NULL);

if (  (*SkipSymbol)
    &&(! IsAncestorOf(*SkipSymbol)))
   { return(FALSE);	/* skip this symbol table (prune the subtree) */
    } /* fi */

/* we don't need to go into UserTypes,   */
/* those symbols cannot depend on others */

Symbol = First();
while(Symbol)
   { if (Symbol == ThatSymbol)
	{ Symbol = Symbol->Succ();
	  continue;	/* nothing depends on itself */
	 } /* fi */
     if (Symbol->DFS_FindDependant(SkipSymbol, ThatSymbol,
					DepSymbol, DepStmnt, DepExpr, Reason))
	{ return(TRUE);
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

return(FALSE);

} /* end of SIR_Symbols::DFS_FindDependant */


void SIR_Symbols::SetAlias(	/* sets all type, usertype, symbol alias' */
	sir_symbols	*Alias)	/* (iterates over symbols and usertypes) */
{
sir_symbol	*Symbol,
		*SymbolAlias;

assert(Alias != NULL);

/* no need to process this node */

if (UserTypes)
   { UserTypes->SetAlias(Alias->UserTypes);
    } /* fi */

Symbol = First();
SymbolAlias = Alias->First();
while(Symbol)
   { Symbol->SetAlias(SymbolAlias);
     Symbol = Symbol->Succ();
     SymbolAlias = SymbolAlias->Succ();
    } /* elihw */

} /* end of SIR_Symbols::SetAlias */


void SIR_Symbols::UnAlias(	/* unalias all type, usertype, symbol links */
	sir_symbols	*GlobalSymbols)
{
sir_symbol	*Symbol;

if (  (Parent)
    &&(Parent->Parent == NULL))
   { Parent = GlobalSymbols;
    } /* fi */
if (  (ParentSymbol)
    &&(ParentSymbol->Alias))
   { ParentSymbol = ParentSymbol->Alias;
    } /* fi */
if (  (ParentUType)
    &&(ParentUType->Alias))
   { ParentUType = ParentUType->Alias;
    } /* fi */

UserTypes->UnAlias(GlobalSymbols);

Symbol = First();
while(Symbol)
   { Symbol->UnAlias(GlobalSymbols);
     Symbol = Symbol->Succ();
    } /* elihw */

} /* end of SIR_Symbols::UnAlias */


ERROR SIR_Symbols::Integrate(		/* integrates imported symbol table */
	sir_symbols	*Imported)
{
sir_symbol	*ImpCurr,
		*ImpSucc,
		*Known;

ImpCurr = Imported->First();	/* for all imported symbols */
while(ImpCurr)
   { ImpSucc = ImpCurr->Succ();
     if ((Known = FindLocally(ImpCurr->Name.chars())))	/* already known? */
	{ /* type and class must match */
	  if (ImpCurr->Class != Known->Class)
	     { SIR_ErrMsg.form("Error while importing:" GL_ERROR_MSG_NEWLINE
				"Redefinition of symbol '%s'",
				ImpCurr->Name.chars());
	       return(SIR_ERROR_REDEFINITION_OF_SYMBOL);
	      } /* fi */
	  if (ImpCurr->Type != Known->Type)	/* types identical? */
	     { SIR_ErrMsg.form(
		"Error while importing:" GL_ERROR_MSG_NEWLINE
		"Declaration of symbol '%s'" GL_ERROR_MSG_NEWLINE
		"does not match former declaration:" GL_ERROR_MSG_NEWLINE
			"type mismatch",
				ImpCurr->Name.chars());
	       return(SIR_ERROR_DECLARATION_MISMATCH_1);
	      } /* fi */
	  /* separate for each class:				*/
	  /* - check matching (eg. storage class)		*/
	  /* - check for redefinition: if yes, complain		*/
	  /* - if no, extend former declaration to a definition	*/
	  switch(ImpCurr->Class)
	     { case SIR_SYMBOL_IDENTIFIER:
		  { if (ImpCurr->IsFunction())	/* is it a function? */
		       { if (ImpCurr->IsFunctionDefinition())
			    { if (Known->IsFunctionDefinition())
				 { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Redefinition of function '%s'",
				ImpCurr->Name.chars());
				   return(SIR_ERROR_REDEFINITION_OF_FUNCTION);
				  } /* fi */
			      /* check storage class */
			      if (  (ImpCurr->StorageClass ==
							SIR_STORAGE_EXTERN)
				  ||(  (ImpCurr->StorageClass !=
						Known->StorageClass)
				     &&(  (ImpCurr->StorageClass !=
							SIR_STORAGE_NONE)
					||(Known->StorageClass !=
							SIR_STORAGE_EXTERN))))
				 { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Definition of function '%s'"
							GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"storage class mismatch",
					ImpCurr->Name.chars());
				   return(SIR_ERROR_SYMBOL_STORAGE_MISMATCH_2);
				  } /* fi */
			      /* make declaration a definition */
			      Known->StorageClass = ImpCurr->StorageClass;
			      assert(Known->Parameters == NULL);
			      Known->Parameters = ImpCurr->Parameters;
			      ImpCurr->Parameters = NULL;
			      assert(Known->ParamScope == NULL);
			      Known->ParamScope = ImpCurr->ParamScope;
			      ImpCurr->ParamScope = NULL;
			      Known->Labels = ImpCurr->Labels;
			      ImpCurr->Labels = NULL;
			      Known->FctBody = ImpCurr->FctBody;
			      ImpCurr->FctBody = NULL;
			      Known->Imported = ImpCurr->Imported;
			      Known->SetLineInfo(ImpCurr);
			     } /* fi */
			 else
			    { /* re-declaration, just check storage class */
			      if (  (ImpCurr->StorageClass !=
						Known->StorageClass)
				  &&(  (ImpCurr->StorageClass !=
							SIR_STORAGE_EXTERN)
				     ||(Known->StorageClass !=
							SIR_STORAGE_NONE)))
				 { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Declaration of function '%s'"
							GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"storage class mismatch",
					ImpCurr->Name.chars());
				   return(SIR_ERROR_SYMBOL_STORAGE_MISMATCH_2);
				  } /* fi */
			     } /* esle */
			} /* fi */
		    else	/* it's a variable or enum constant */
				/* (we are on top-level here, so it */
				/* cannot be a port or instance)    */
		       { assert(  (ImpCurr->IsVariable())
				||(ImpCurr->IsEnumMember()));
			 if (  (ImpCurr->IsVariableDefinition())
			     ||(ImpCurr->IsEnumMember()))
			    { if (  (Known->IsVariableDefinition())
				  ||(Known->IsEnumMember()))
				 { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Symbol '%s' defined twice",
				ImpCurr->Name.chars());
				   return(SIR_ERROR_SYMBOL_DEFINED_TWICE_1);
				  } /* fi */
			      /* check storage class */
			      if (  (ImpCurr->StorageClass !=
						Known->StorageClass)
				  &&(  (ImpCurr->StorageClass !=
							SIR_STORAGE_NONE)
				     ||(Known->StorageClass !=
							SIR_STORAGE_EXTERN)))
				 { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Definition of symbol '%s'" GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"storage class mismatch",
					ImpCurr->Name.chars());
				   return(SIR_ERROR_SYMBOL_STORAGE_MISMATCH_2);
				  } /* fi */
			      /* make declaration a definition */
			      Known->StorageClass = ImpCurr->StorageClass;
			      assert(ImpCurr->PipeStages == 0);
			      assert(Known->Initializer == NULL);
			      Known->Initializer = ImpCurr->Initializer;
			      ImpCurr->Initializer = NULL;
			      Known->Imported = ImpCurr->Imported;
			      Known->SetLineInfo(ImpCurr);
			     } /* fi */
			 else
			    { /* re-declaration, just check storage class */
			      if (  (ImpCurr->StorageClass !=
						Known->StorageClass)
				  &&(  (ImpCurr->StorageClass !=
							SIR_STORAGE_EXTERN)
				     ||(Known->StorageClass !=
							SIR_STORAGE_NONE)))
				 { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Declaration of symbol '%s'"
							GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"storage class mismatch",
					ImpCurr->Name.chars());
				   return(SIR_ERROR_SYMBOL_STORAGE_MISMATCH_2);
				  } /* fi */
			     } /* esle */
			} /* esle */
		    break;
		   }
	       case SIR_SYMBOL_TYPEDEF:
		  { break;	/* nothing else to check or to extend */
		   }
	       case SIR_SYMBOL_BEHAVIOR:
		  { /* implemented interfaces must match (with names) */
		    assert(Known->Interfaces != NULL);
		    assert(ImpCurr->Interfaces != NULL);
		    if (0 != Known->Interfaces->CmpSymbolNames(
						ImpCurr->Interfaces))
		       { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Declaration of behavior '%s'"
							GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"mismatch of implemented interfaces",
				ImpCurr->Name.chars());
			 return(SIR_ERROR_DECLARATION_MISMATCH_2);
			} /* fi */
		    if (ImpCurr->IsBehaviorDefinition())
		       { if (Known->IsBehaviorDefinition())
			    { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Redefinition of behavior '%s'",
				ImpCurr->Name.chars());
			      return(SIR_ERROR_REDEFINITION_OF_BEHAVIOR);
			     } /* fi */
			 /* make declaration a definition */
			 assert(Known->Parameters == NULL);
			 Known->Parameters = ImpCurr->Parameters;
			 ImpCurr->Parameters = NULL;
			 assert(Known->ParamScope == NULL);
			 Known->ParamScope = ImpCurr->ParamScope;
			 ImpCurr->ParamScope = NULL;
			 Known->ClassScope = ImpCurr->ClassScope;
			 ImpCurr->ClassScope = NULL;
			 Known->Imported = ImpCurr->Imported;
			 Known->SetLineInfo(ImpCurr);
			} /* fi */
		    else
		       { /* re-declaration, nothing to do */
			} /* esle */
		    break;
		   }
	       case SIR_SYMBOL_CHANNEL:
		  { /* implemented interfaces must match (with names) */
		    assert(Known->Interfaces != NULL);
		    assert(ImpCurr->Interfaces != NULL);
		    if (0 != Known->Interfaces->CmpSymbolNames(
						ImpCurr->Interfaces))
		       { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Declaration of channel '%s'"
							GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"mismatch of implemented interfaces",
				ImpCurr->Name.chars());
			 return(SIR_ERROR_DECLARATION_MISMATCH_2);
			} /* fi */
		    if (ImpCurr->IsChannelDefinition())
		       { if (Known->IsChannelDefinition())
			    { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Redefinition of channel '%s'",
				ImpCurr->Name.chars());
			      return(SIR_ERROR_REDEFINITION_OF_CHANNEL);
			     } /* fi */
			 /* make declaration a definition */
			 assert(Known->Parameters == NULL);
			 Known->Parameters = ImpCurr->Parameters;
			 ImpCurr->Parameters = NULL;
			 assert(Known->ParamScope == NULL);
			 Known->ParamScope = ImpCurr->ParamScope;
			 ImpCurr->ParamScope = NULL;
			 Known->ClassScope = ImpCurr->ClassScope;
			 ImpCurr->ClassScope = NULL;
			 Known->Imported = ImpCurr->Imported;
			 Known->SetLineInfo(ImpCurr);
			} /* fi */
		    else
		       { /* re-declaration, nothing to do */
			} /* esle */
		    break;
		   }
	       case SIR_SYMBOL_INTERFACE:
		  { if (ImpCurr->IsInterfaceDefinition())
		       { if (Known->IsInterfaceDefinition())
			    { SIR_ErrMsg.form(
				"Error while importing:" GL_ERROR_MSG_NEWLINE
				"Redefinition of interface '%s'",
				ImpCurr->Name.chars());
			      return(SIR_ERROR_REDEFINITION_OF_INTERFACE);
			     } /* fi */
			 /* make declaration a definition */
			 Known->ClassScope = ImpCurr->ClassScope;
			 ImpCurr->ClassScope = NULL;
			 Known->Imported = ImpCurr->Imported;
			 Known->SetLineInfo(ImpCurr);
			} /* fi */
		    else
		       { /* re-declaration, nothing to do */
			} /* esle */
		    break;
		   }
	       default:
		  { assert(FALSE);	/* bad symbol class */
		   }
	      } /* hctiws */
	  /* set the alias pointer (in all these cases) */
	  ImpCurr->Alias = Known;
	 } /* fi */
     else	/* symbol is not known yet */
	{ /* move the complete symbol into the actual symbol table */
	  if (Curr())
	     { InsertBefore(Imported->Remove(ImpCurr));
	      } /* fi */
	  else
	     { Append(Imported->Remove(ImpCurr));
	      } /* esle */
	 } /* esle */
     ImpCurr = ImpSucc;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::Integrate */


ERROR SIR_Symbols::CheckClassContainment(void)	/* checks class hierarchy */
{						/* for cycles */
sir_symbol	*Symbol;

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsBehaviorDefinition())
	 ||(Symbol->IsChannelDefinition()))
	{ if (Symbol->Contains(Symbol))	/* contains itself? */
	     { SIR_ErrMsg.form(
		"Error while importing:" GL_ERROR_MSG_NEWLINE
		"class containment cycle detected;" GL_ERROR_MSG_NEWLINE
		"class '%s' contains itself",
			Symbol->Name.chars());
	       return(SIR_ERROR_CLASS_CONTAINMENT_CYCLE);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::CheckClassContainment */


sir_symbol *SIR_Symbols::Find(	/* find an entry globally (here and above) */
	const char	*Name)	/* (returns NULL if not found) */
{
sir_symbol	*Symbol;

if ((Symbol = FindLocally(Name)))
   { return(Symbol);
    } /* fi */

if (Parent)
   { return(Parent->Find(Name));
    } /* fi */
else
   { return(NULL);	/* not found */
    } /* esle */

} /* end of SIR_Symbols::Find */


sir_symbol *SIR_Symbols::FindLocally(	/* find an entry locally */
	const char	*Name)		/* (returns NULL if not found) */
{
int		CmpVal;

First();
while(Curr())
   { if (0 == (CmpVal = strcmp(Name, Curr()->Name.chars())))
	{ return(Curr());
	 } /* fi */
     if (CmpVal < 0)
	{ break;
	 } /* fi */
     Next();
    } /* elihw */

return(NULL);	/* not found */

} /* end of SIR_Symbols::FindLocally */


sir_symbol *SIR_Symbols::FindLocally(	/* find a symbol with scope locally */
	sir_symbols	*Scope)		/* (returns NULL if not found) */
{

First();
while(Curr())
   { if (Scope == Curr()->ParamScope)
	{ return(Curr());
	 } /* fi */
     if (Scope == Curr()->ClassScope)
	{ return(Curr());
	 } /* fi */
     Next();
    } /* elihw */

return(NULL);	/* not found */

} /* end of SIR_Symbols::FindLocally */


sir_symbol *SIR_Symbols::Insert(	/* inserts a new symbol */
	sir_symbol	*NewSymbol)
{
int		CmpVal;

First();
while(Curr())
   { if (0 == (CmpVal = strcmp(NewSymbol->Name.chars(),
				Curr()->Name.chars())))
	{ assert(FALSE);	/* already exists! */
	 } /* fi */
     if (CmpVal < 0)
	{ break;
	 } /* fi */
     Next();
    } /* elihw */

if (Curr())
   { return(InsertBefore(NewSymbol));
    } /* fi */
else
   { return(Append(NewSymbol));
    } /* esle */

} /* end of SIR_Symbols::Insert */


void SIR_Symbols::FindName(		/* find a symbol or usertype globally */
	const char	*Name,
	sir_symbol	**SymbolFound,		/* the symbol found (or NULL) */
	sir_usertype	**UserTypeFound)	/* user-type found (or NULL) */
{

FindNameLocally(Name, SymbolFound, UserTypeFound);

if ((*SymbolFound) || (*UserTypeFound))
   { return;
    } /* fi */

if (Parent)
   { Parent->FindName(Name, SymbolFound, UserTypeFound);
    } /* fi */

} /* end of SIR_Symbols::FindName */


void SIR_Symbols::FindNameLocally(	/* find a symbol or usertype locally */
	const char	*Name,
	sir_symbol	**SymbolFound,		/* the symbol found (or NULL) */
	sir_usertype	**UserTypeFound)	/* user-type found (or NULL) */
{

if ((*SymbolFound = FindLocally(Name)))
   { *UserTypeFound = NULL;
     return;
    } /* fi */

*UserTypeFound = UserTypes->FindLocally(Name);

} /* end of SIR_Symbols::FindNameLocally */


sir_symbol *SIR_Symbols::Declare(/* declares a new symbol (and prepares def.) */
	sir_symbol	*NewSymbol)	/* (the argument is consumed!) */
{					/* (may return NULL and SIR_Error) */
sir_symbol	*OldSymbol,
		*HiddenSymbol,
		*Dependant;

assert(NewSymbol != NULL);		/* note: this method is called */
assert(NewSymbol->LineInfo != NULL);	/* exclusively from the parser */

/* first we check the StorageClass depending on symbol class and this scope */

if ((SIR_Error = NewSymbol->CheckStorageClass(
					&NewSymbol->StorageClass,
					&NewSymbol->PipeStages,
					&NewSymbol->Class,
					&ScopeInfo)))
   { delete NewSymbol;
     return(NULL);
    } /* fi */

/* then we check special requirements for behaviors and channels */

switch(NewSymbol->Class)
   { case SIR_SYMBOL_IDENTIFIER:
     case SIR_SYMBOL_TYPEDEF:
     case SIR_SYMBOL_INTERFACE:
	{ break;
	 }
     case SIR_SYMBOL_BEHAVIOR:
	{ if (0 == strcmp(NewSymbol->Name.chars(), GL_MAIN_BEHAVIOR_NAME))
	     { if (! NewSymbol->Type->VoidParameters())
		  { SIR_Error = SIR_ERROR_NO_PORTS_ALLOWED_FOR_MAIN;
		    delete NewSymbol;
		    return(NULL);
		   } /* fi */
	       assert(NewSymbol->Interfaces != NULL);	/* is a behavior */
	       if (! NewSymbol->Interfaces->Empty())
		  { SIR_Error = SIR_ERROR_NO_INTERFACES_ALLOWED_FR_MAIN;
		    delete NewSymbol;
		    return(NULL);
		   } /* fi */
	      } /* fi */
	  break;
	 }
     case SIR_SYMBOL_CHANNEL:
	{ assert(NewSymbol->Type->Parameters != NULL);
	  assert(NewSymbol->Interfaces != NULL);
	  if (  (NewSymbol->Type->VoidParameters())
	      &&(NewSymbol->Interfaces->Empty()))
	     { GL_PrintWarningFmt(GL_WARN_STANDARD,
		"Channel '%s' is useless because it has" GL_WARN_MSG_NEWLINE
		"no ports and no implemented interfaces",
			NewSymbol->Name.chars());
	      } /* fi */
	  break;
	 }
     default:
	{ assert(FALSE);	/* bad symbol class */
	 }
    } /* hctiws */

#ifdef SIR_WARN_ABOUT_SHADOWED_SYMBOLS
if (  (ScopeInfo != SIR_SCOPE_USERTYPE)
    &&(Parent)
    &&(!(  (ScopeInfo == SIR_SCOPE_PARAMETER)		// bug fix 06/18/00, RD
	 &&(  (Parent->ScopeInfo == SIR_SCOPE_USERTYPE)	// bug fix 06/18/00, RD
	    ||(Parent->ScopeInfo == SIR_SCOPE_PARAMETER))))	// 06/21/00, RD
    &&(HiddenSymbol = Parent->Find(NewSymbol->Name.chars())))
   { if (HiddenSymbol->LineInfo)
	{ GL_PrintWarningFmt(GL_WARN_STANDARD,
		"In line %u, file \"%s\"," GL_WARN_MSG_NEWLINE
		"declaration of symbol '%s' hides symbol '%s'"
							GL_WARN_MSG_NEWLINE
		"declared in line %u, file \"%s\"",
			NewSymbol->LineInfo->Line,
			NewSymbol->LineInfo->File->Filename.chars(),
			NewSymbol->Name.chars(),
			HiddenSymbol->Name.chars(),
			HiddenSymbol->LineInfo->Line,
			HiddenSymbol->LineInfo->File->Filename.chars());
	 } /* fi */
     else
	{ GL_PrintWarningFmt(GL_WARN_STANDARD,
		"In line %u, file \"%s\"," GL_WARN_MSG_NEWLINE
		"declaration of symbol '%s' hides symbol '%s'"
							GL_WARN_MSG_NEWLINE
		"declared at higher level",
			NewSymbol->LineInfo->Line,
			NewSymbol->LineInfo->File->Filename.chars(),
			NewSymbol->Name.chars(),
			HiddenSymbol->Name.chars());
	 } /* esle */
    } /* fi */
#endif /* SIR_WARN_ABOUT_SHADOWED_SYMBOLS */

/* then we check for scoping problems in classes */

if (ScopeInfo == SIR_SCOPE_CLASS)
   { assert(Parent != NULL);
     if ((HiddenSymbol = Parent->Find(NewSymbol->Name.chars())))
	{ if (HiddenSymbol->FindDependant(&Dependant, NULL, NULL, NULL, this))
	     { SIR_ErrMsg.form(
		"Declaration of class member/method '%s'" GL_ERROR_MSG_NEWLINE
		"hides global symbol '%s' from being" GL_ERROR_MSG_NEWLINE
		"referenced by local symbol '%s'",
			NewSymbol->Name.chars(),
			HiddenSymbol->Name.chars(),
			Dependant->Name.chars());
	       SIR_Error = SIR_ERROR_MEMBER_HIDES_USED_GLOBAL_SYMB;
	       delete NewSymbol;
	       return(NULL);
	      } /* fi */
	 } /* fi */
    } /* fi */

/* then we check class instantiations and their connections */

if (  (NewSymbol->IsBehaviorInstance())	/* check class instantiations */
    ||(NewSymbol->IsChannelInstance()))
   {
     /* first, make sure that the hierarchy will not have any cycles */

     assert(ParentSymbol != NULL);
     if (  (NewSymbol->Type->ClassSymbol == ParentSymbol) /* self-inst.? */
	 ||(NewSymbol->Type->ClassSymbol->Contains(ParentSymbol)))
	{ SIR_ErrMsg.form(
		"Cyclic instantiation: Class '%s' contains" GL_ERROR_MSG_NEWLINE
		"instances of class '%s'",
		NewSymbol->Type->ClassSymbol->Name.chars(),
		ParentSymbol->Name.chars());
	  SIR_Error = SIR_ERROR_CYCLIC_INSTANTIATION;
	  delete NewSymbol;
	  return(NULL);
	 } /* fi */

     /* next we check the port mapping list semantically */
     /* (note: this solves BUGS_V200/BUG5 and partially BUGS_V200/BUG4) */

     assert(NewSymbol->PortMappings != NULL);
     assert(NewSymbol->Type->Parameters != NULL);
     if ((SIR_Error = NewSymbol->PortMappings->Check(
				NewSymbol->Type->Parameters,
				NewSymbol->LineInfo->File->Filename.chars(),
				NewSymbol->LineInfo->Line)))
	{ delete NewSymbol;
	  return(NULL);
	 } /* fi */
    } /* fi */

/* everything looks ok, now we can declare the symbol */

if ((OldSymbol = FindLocally(NewSymbol->Name.chars())))	/* already there? */
   { /* type and class must match */
     if (NewSymbol->Class != OldSymbol->Class)
	{ SIR_ErrMsg.form("Redefinition of symbol '%s'",
				NewSymbol->Name.chars());
	  SIR_Error = SIR_ERROR_REDEFINITION_OF_SYMBOL_2;
	  delete NewSymbol;
	  return(NULL);
	 } /* fi */
     if (NewSymbol->Type != OldSymbol->Type)	/* types identical? */
	{ SIR_ErrMsg.form(
		"Declaration of symbol '%s'" GL_ERROR_MSG_NEWLINE
		"does not match former declaration:" GL_ERROR_MSG_NEWLINE
		"type mismatch",
		NewSymbol->Name.chars());
	  SIR_Error = SIR_ERROR_DECLARATION_MISMATCH_3;
	  delete NewSymbol;
	  return(NULL);
	 } /* fi */
     /* separate for each class:				*/
     /* - check matching (eg. storage class), else complain	*/
     /* - check for redefinition: if yes, complain		*/
     /* - if no, prepare former declaration for definition	*/
     /*   with DefineXxxx() methods				*/
     switch(NewSymbol->Class)
	{ case SIR_SYMBOL_IDENTIFIER:
	     { if (NewSymbol->IsFunction())	/* is it a function? */
		  { if (NewSymbol->Parameters)	/* prepare definition? */
		       { if (OldSymbol->IsFunctionDefinition())
			    { SIR_ErrMsg.form(
				"Redefinition of function '%s'",
				NewSymbol->Name.chars());
			      SIR_Error = SIR_ERROR_REDEFINITION_OF_FUNCTION_2;
			      delete NewSymbol;
			      return(NULL);
			     } /* fi */
			 /* check storage class */
			 if (  (NewSymbol->StorageClass == SIR_STORAGE_EXTERN)
			     ||(  (NewSymbol->StorageClass !=
						OldSymbol->StorageClass)
				&&(  (NewSymbol->StorageClass !=
							SIR_STORAGE_NONE)
				   ||(OldSymbol->StorageClass !=
							SIR_STORAGE_EXTERN))))
			    { SIR_ErrMsg.form(
				"Definition of function '%s'"
							GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"storage class mismatch",
				NewSymbol->Name.chars());
			      SIR_Error = SIR_ERROR_SYMBOL_STORAGE_MISMATCH_1;
			      delete NewSymbol;
			      return(NULL);
			     } /* fi */
			 /* make declaration a prepared definition */
			 OldSymbol->StorageClass = NewSymbol->StorageClass;
			 assert(OldSymbol->PipeStages == 0);
			 assert(NewSymbol->PipeStages == 0);
			 assert(OldSymbol->Parameters == NULL);
			 OldSymbol->Parameters = NewSymbol->Parameters;
			 NewSymbol->Parameters = NULL;
			 assert(OldSymbol->ParamScope == NULL);
			 assert(NewSymbol->ParamScope != NULL);
			 OldSymbol->ParamScope = NewSymbol->ParamScope;
			 NewSymbol->ParamScope = NULL;
			 OldSymbol->ParamScope->ParentSymbol = OldSymbol;
			 assert(OldSymbol->Labels == NULL);
			 assert(NewSymbol->Labels != NULL);
			 OldSymbol->Labels = NewSymbol->Labels;
			 NewSymbol->Labels = NULL;
			 /* FctBody is handled by DefineFunction() */
			 OldSymbol->Imported = NewSymbol->Imported;
			 OldSymbol->SetLineInfo(NewSymbol);
			} /* fi */
		    else
		       { /* re-declaration, just check storage class */
			 if (  (NewSymbol->StorageClass !=
						OldSymbol->StorageClass)
			     &&(  (NewSymbol->StorageClass !=
							SIR_STORAGE_EXTERN)
				||(OldSymbol->StorageClass !=
							SIR_STORAGE_NONE)))
			    { SIR_ErrMsg.form(
				"Declaration of function '%s'"
							GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"storage class mismatch",
				NewSymbol->Name.chars());
			      SIR_Error = SIR_ERROR_SYMBOL_STORAGE_MISMATCH_1;
			      delete NewSymbol;
			      return(NULL);
			     } /* fi */
			} /* esle */
		   } /* fi */
	       else	/* it's a variable, enum constant, instance, or port */
		  { assert(  (NewSymbol->IsVariable())
			   ||(NewSymbol->IsEnumMember())
			   ||(NewSymbol->IsBehaviorInstance())
			   ||(NewSymbol->IsChannelInstance())
			   ||(NewSymbol->IsPort()));
		    if (  (NewSymbol->IsVariableDefinition())
			||(NewSymbol->IsEnumMember())
			||(NewSymbol->IsBehaviorInstance())
			||(NewSymbol->IsChannelInstance())
			||(NewSymbol->IsPort()))
		       { if (  (OldSymbol->IsVariableDefinition())
			     ||(OldSymbol->IsEnumMember())
			     ||(OldSymbol->IsBehaviorInstance())
			     ||(OldSymbol->IsChannelInstance())
			     ||(OldSymbol->IsPort()))
			    { SIR_ErrMsg.form("Symbol '%s' defined twice",
						NewSymbol->Name.chars());
			      SIR_Error = SIR_ERROR_SYMBOL_DEFINED_TWICE_2;
			      delete NewSymbol;
			      return(NULL);
			     } /* fi */
			 /* check storage class */
			 if (  (NewSymbol->StorageClass !=
						OldSymbol->StorageClass)
			     &&(  (NewSymbol->StorageClass !=
							SIR_STORAGE_NONE)
				||(OldSymbol->StorageClass !=
							SIR_STORAGE_EXTERN)))
			    { SIR_ErrMsg.form(
				"Definition of symbol '%s'"
							GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"storage class mismatch",
				NewSymbol->Name.chars());
			      SIR_Error = SIR_ERROR_SYMBOL_STORAGE_MISMATCH_1;
			      delete NewSymbol;
			      return(NULL);
			     } /* fi */
			 /* make declaration a prepared definition */
			 OldSymbol->StorageClass = NewSymbol->StorageClass;
			 assert(OldSymbol->PipeStages == 0);
			 assert(NewSymbol->PipeStages == 0);
			 /* Initializer is handled by DefineVariable() */
			 OldSymbol->Imported = NewSymbol->Imported;
			 OldSymbol->SetLineInfo(NewSymbol);
			} /* fi */
		    else
		       { /* re-declaration, just check storage class */
			 if (  (NewSymbol->StorageClass !=
						OldSymbol->StorageClass)
			     &&(  (NewSymbol->StorageClass !=
							SIR_STORAGE_EXTERN)
				||(OldSymbol->StorageClass !=
							SIR_STORAGE_NONE)))
			    { SIR_ErrMsg.form(
				"Declaration of symbol '%s'"
							GL_ERROR_MSG_NEWLINE
				"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
				"storage class mismatch",
				NewSymbol->Name.chars());
			      SIR_Error = SIR_ERROR_SYMBOL_STORAGE_MISMATCH_1;
			      delete NewSymbol;
			      return(NULL);
			     } /* fi */
			} /* esle */
		   } /* esle */
	       delete NewSymbol;
	       return(OldSymbol);	/* return (updated) old symbol */
	      }
	  case SIR_SYMBOL_TYPEDEF:
	     { /* this is always a declaration, */
	       /* but we may have to set the symbol link for usertypes */
	       if (OldSymbol->Type->UserType)	/* is a usertype? */
		  { if (OldSymbol->Type->UserType->TypeDef == NULL)
		       { OldSymbol->Type->UserType->TypeDef = OldSymbol;
			} /* fi */
		   } /* fi */
	       delete NewSymbol;
	       return(OldSymbol);	/* return old symbol */
	      }
	  case SIR_SYMBOL_BEHAVIOR:
	     { /* implemented interfaces must match (with names) */
	       assert(OldSymbol->Interfaces != NULL);
	       assert(NewSymbol->Interfaces != NULL);
	       if (0 != OldSymbol->Interfaces->CmpSymbolNames(
						NewSymbol->Interfaces))
		  { SIR_ErrMsg.form(
			"Declaration of behavior '%s'" GL_ERROR_MSG_NEWLINE
			"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
			"mismatch of implemented interfaces",
			NewSymbol->Name.chars());
		    SIR_Error = SIR_ERROR_DECLARATION_MISMATCH_4;
		    delete NewSymbol;
		    return(NULL);
		   } /* fi */
	       if (NewSymbol->Parameters)	/* prepare definition? */
		  { if (OldSymbol->IsBehaviorDefinition())
		       { SIR_ErrMsg.form("Redefinition of behavior '%s'",
						NewSymbol->Name.chars());
			 SIR_Error = SIR_ERROR_REDEFINITION_OF_BEHAVIOR_2;
			 delete NewSymbol;
			 return(NULL);
			} /* fi */
		    /* make declaration a prepared definition */
		    assert(OldSymbol->Parameters == NULL);
		    OldSymbol->Parameters = NewSymbol->Parameters;
		    NewSymbol->Parameters = NULL;
		    assert(OldSymbol->ClassScope == NULL);
		    assert(NewSymbol->ClassScope != NULL);
		    OldSymbol->ClassScope = NewSymbol->ClassScope;
		    NewSymbol->ClassScope = NULL;
		    OldSymbol->ClassScope->ParentSymbol = OldSymbol;
		    OldSymbol->Imported = NewSymbol->Imported;
		    OldSymbol->SetLineInfo(NewSymbol);
		   } /* fi */
	       else
		  { /* re-declaration, nothing to do */
		   } /* esle */
	       delete NewSymbol;
	       return(OldSymbol);	/* return (updated) old behavior */
	      }
	  case SIR_SYMBOL_CHANNEL:
	     { /* implemented interfaces must match (with names) */
	       assert(OldSymbol->Interfaces != NULL);
	       assert(NewSymbol->Interfaces != NULL);
	       if (0 != OldSymbol->Interfaces->CmpSymbolNames(
						NewSymbol->Interfaces))
		  { SIR_ErrMsg.form(
			"Declaration of channel '%s'" GL_ERROR_MSG_NEWLINE
			"does not match former declaration:"
							GL_ERROR_MSG_NEWLINE
			"mismatch of implemented interfaces",
			NewSymbol->Name.chars());
		    SIR_Error = SIR_ERROR_DECLARATION_MISMATCH_4;
		    delete NewSymbol;
		    return(NULL);
		   } /* fi */
	       if (NewSymbol->Parameters)	/* prepare definition? */
		  { if (OldSymbol->IsChannelDefinition())
		       { SIR_ErrMsg.form("Redefinition of channel '%s'",
						NewSymbol->Name.chars());
			 SIR_Error = SIR_ERROR_REDEFINITION_OF_CHANNEL_2;
			 delete NewSymbol;
			 return(NULL);
			} /* fi */
		    /* make declaration a prepared definition */
		    assert(OldSymbol->Parameters == NULL);
		    OldSymbol->Parameters = NewSymbol->Parameters;
		    NewSymbol->Parameters = NULL;
		    assert(OldSymbol->ClassScope == NULL);
		    assert(NewSymbol->ClassScope != NULL);
		    OldSymbol->ClassScope = NewSymbol->ClassScope;
		    NewSymbol->ClassScope = NULL;
		    OldSymbol->ClassScope->ParentSymbol = OldSymbol;
		    OldSymbol->Imported = NewSymbol->Imported;
		    OldSymbol->SetLineInfo(NewSymbol);
		   } /* fi */
	       else
		  { /* re-declaration, nothing to do */
		   } /* esle */
	       delete NewSymbol;
	       return(OldSymbol);	/* return (updated) old channel */
	      }
	  case SIR_SYMBOL_INTERFACE:
	     { if (NewSymbol->ClassScope)	/* prepare definition? */
		  { if (OldSymbol->IsInterfaceDefinition())
		       { SIR_ErrMsg.form("Redefinition of interface '%s'",
						NewSymbol->Name.chars());
			 SIR_Error = SIR_ERROR_REDEFINITION_OF_INTERFACE_2;
			 delete NewSymbol;
			 return(NULL);
			} /* fi */
		    /* make declaration a prepared definition */
		    assert(OldSymbol->ClassScope == NULL);
		    assert(NewSymbol->ClassScope != NULL);
		    OldSymbol->ClassScope = NewSymbol->ClassScope;
		    NewSymbol->ClassScope = NULL;
		    OldSymbol->ClassScope->ParentSymbol = OldSymbol;
		    OldSymbol->Imported = NewSymbol->Imported;
		    OldSymbol->SetLineInfo(NewSymbol);
		   } /* fi */
	       else
		  { /* re-declaration, nothing to do */
		   } /* esle */
	       delete NewSymbol;
	       return(OldSymbol);	/* return old interface */
	      }
	  default:
	     { assert(FALSE);	/* bad symbol class */
	      }
	 } /* hctiws */
    } /* fi */

/* not found, so we just insert it */
assert(OldSymbol == NULL);

/* but before we may have to set the typedef link for (unnamed) usertypes */
if (  (NewSymbol->Type->UserType)			/* is a usertype? */
    &&(NewSymbol->StorageClass == SIR_STORAGE_TYPEDEF))	/* and a typedef? */
   { if (NewSymbol->Type->UserType->TypeDef == NULL)	/* is still unset? */
	{ NewSymbol->Type->UserType->TypeDef = NewSymbol;
	 } /* fi */
    } /* fi */

/* and for classes we have to fix pointers from the symbols type */
if (  (NewSymbol->Class == SIR_SYMBOL_BEHAVIOR)
    ||(NewSymbol->Class == SIR_SYMBOL_CHANNEL)
    ||(NewSymbol->Class == SIR_SYMBOL_INTERFACE))
   { NewSymbol->Type->ClassSymbol = NewSymbol;
     NewSymbol->Type->Name = NewSymbol->Name.chars();
    } /* fi */

/* FindLocally() already located the insertion spot */
if (Curr())
   { return(InsertBefore(NewSymbol));
    } /* fi */
else
   { return(Append(NewSymbol));
    } /* esle */

} /* end of SIR_Symbols::Declare */


ERROR SIR_Symbols::DefineVariable(	/* makes variable decl. a definition */
	sir_symbol	*VariableDecl,	/* (consumes Initializer) */
	SIR_STORAGE	DeclStorageClass,
	sir_initializer	*Initializer)
{

assert(VariableDecl != NULL);

if (Initializer)
   { assert(VariableDecl->Initializer == NULL);	/* not defined yet */
     if ((SIR_Error = VariableDecl->CheckInitializer(
					Initializer,
					NULL,
					&DeclStorageClass)))
	{ delete Initializer;
	  return(SIR_Error);
	 } /* fi */
     VariableDecl->Initializer = Initializer;	/* store initializer */
     VariableDecl->Type->ConvertInitializer(	/* convert to expected type */
				Initializer);
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::DefineVariable */


ERROR SIR_Symbols::DefineFunction(	/* makes function decl. a definition */
	sir_symbol	*FunctionDecl,	/* (consumes FctBody) */
	sir_statement	*FctBody)
{

assert(FunctionDecl != NULL);
assert(FunctionDecl->FctBody == NULL);	/* not defined yet */
assert(FctBody != NULL);
assert(FunctionDecl->Labels != NULL);

FunctionDecl->FctBody = FctBody;	/* store function body */

if ((SIR_Error = FunctionDecl->Labels->CheckDefStatements())) /* check labels */
   { return(SIR_Error);
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::DefineFunction */


ERROR SIR_Symbols::DefineClass(		/* makes class decl. a definition */
	sir_symbol	*ClassDecl)
{
sir_symbol	*MainMethod,
		*Symbol;
sir_symbol_ptr	*Interface;
sir_symbol	*Declared,
		*Implemented;

assert(ClassDecl != NULL);
assert(ClassDecl->ClassScope != NULL);

/* first check if the name reserved for constructor/destructor is used */

if (ClassDecl->ClassScope->FindLocally(ClassDecl->Name.chars()))
   { SIR_ErrMsg.form("Name '%s' is reserved in its class" GL_ERROR_MSG_NEWLINE
		"and cannot be redeclared in its class scope",
			ClassDecl->Name.chars());
     return(SIR_ERROR_RESERVED_SYMBOL_NAME);
    } /* fi */

if (ClassDecl->IsBehavior())	/* behaviors must implement a main method */
   { if (!(MainMethod = ClassDecl->ClassScope->FindLocally(
						GL_MAIN_METHOD_NAME)))
	{ SIR_ErrMsg.form(
		"No " GL_MAIN_METHOD_NAME " method declared in behavior '%s'",
			ClassDecl->Name.chars());
	  return(SIR_ERROR_NO_MAIN_METHOD);
	 } /* fi */
     if (! MainMethod->IsFunction())
	{ SIR_ErrMsg.form(
		"Type mismatch in behavior '%s':" GL_ERROR_MSG_NEWLINE
		"member '" GL_MAIN_METHOD_NAME "' is not a method",
			ClassDecl->Name.chars());
	  return(SIR_ERROR_MEMBER_MAIN_NOT_A_METHOD);
	 } /* fi */
     if (  (0 != strcmp(ClassDecl->Name.chars(), GL_MAIN_BEHAVIOR_NAME))
	 &&(  (MainMethod->Type->SubType->Type != SIR_TYPE_VOID)
	    ||(! MainMethod->Type->VoidParameters())))
	{ SIR_ErrMsg.form(
		"Type mismatch in behavior '%s':" GL_ERROR_MSG_NEWLINE
		"method '" GL_MAIN_METHOD_NAME "' is not 'void main(void)'",
			ClassDecl->Name.chars());
	  return(SIR_ERROR_MAIN_METHOD_TYPE_MISMATCH);
	 } /* fi */
    } /* fi */

if (  (ClassDecl->IsBehavior())	/* behaviors and channels must implement */
    ||(ClassDecl->IsChannel()))	/* the listed interfaces */
   { assert(ClassDecl->Interfaces != NULL);
     Interface = ClassDecl->Interfaces->First();
     while(Interface)	/* in each listed interface */
	{ if (! Interface->Symbol->IsInterfaceDefinition())
	     { SIR_ErrMsg.form(
			"Class '%s' implements incomplete interface '%s'",
					ClassDecl->Name.chars(),
					Interface->Symbol->Name.chars());
	       return(SIR_ERROR_IMPLEMENTS_INCOMPLETE_IF);
	      } /* fi */
	  Declared = Interface->Symbol->ClassScope->First();
	  while(Declared)	/* check each declared method */
	     { assert(Declared->IsFunction());	/* nothing else allowed */
	       if (!(Implemented = ClassDecl->ClassScope->FindLocally(
						Declared->Name.chars())))
		  { SIR_ErrMsg.form(
			"Class '%s' lacks implementation of method"
							GL_ERROR_MSG_NEWLINE
			"'%s' declared in interface '%s'",
					ClassDecl->Name.chars(),
					Declared->Name.chars(),
					Interface->Symbol->Name.chars());
		    return(SIR_ERROR_CLASS_LACKS_IMPLEMENTATION);
		   } /* fi */
	       if (!(Declared->Type == Implemented->Type))
		  { SIR_ErrMsg.form(
			"Type mismatch in class '%s':" GL_ERROR_MSG_NEWLINE
			"method '%s' declared in interface '%s'"
							GL_ERROR_MSG_NEWLINE
			"does not match the implementation",
					ClassDecl->Name.chars(),
					Declared->Name.chars(),
					Interface->Symbol->Name.chars());
		    return(SIR_ERROR_IMPLEMENTATION_TYPE_MISMATCH);
		   } /* fi */
	       Declared = Declared->Succ();
	      } /* elihw */
	  Interface = Interface->Succ();
	 } /* elihw */
    } /* fi */

if (ClassDecl->IsInterface())	/* interfaces have restrictions */
   { Symbol = ClassDecl->ClassScope->First();
     while(Symbol)
	{ if (! Symbol->IsFunction())
	     { SIR_ErrMsg.form(
			"Interface member '%s' is not a function declaration",
				Symbol->Name.chars());
	       return(SIR_ERROR_ILLEGAL_INTERFACE_MEMBER);
	      } /* fi */
	  assert(! Symbol->IsFunctionDefinition()); /* grammar assures this */
	  Symbol = Symbol->Succ();
	 } /* elihw */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::DefineClass */


ERROR SIR_Symbols::WriteSC(	/* (re-) generates SpecC source code */
	FILE		*File,
	BOOL		WriteNotes)
{
sir_symbol	*Symbol;

/* (1)-(4) User-defined types, declarations and definitions */

if (UserTypes)
   { if ((SIR_Error = UserTypes->WriteSC(File, WriteNotes, FALSE)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */

/* (5) Type definitions (alias types) */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (5) type definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsTypeDef())
	 &&((  (Symbol->Type->UserType == NULL)	/* except unnamed usertypes */
	     ||(Symbol->Type->UserType->Name != NULL))))
	{ if ((SIR_Error = Symbol->WriteSC(File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (6) Interface declarations */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (6) interface declarations", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsInterface())
	{ if ((SIR_Error = Symbol->WriteSC2(File, WriteNotes, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (7) Channel declarations */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (7) channel declarations", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsChannel())
	{ if ((SIR_Error = Symbol->WriteSC2(File, WriteNotes, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (8) Behavior declarations */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (8) behavior declarations", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsBehavior())
	{ if ((SIR_Error = Symbol->WriteSC2(File, WriteNotes, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (9) Declarations of variables and functions */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (9) declarations of variables and functions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsVariable())
	 ||(Symbol->IsFunction()))
	{ if ((SIR_Error = Symbol->WriteSC(File, WriteNotes, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (10) Variable definitions and class instantiations (SORTED) */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (10a) variable definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsVariableDefinition())
	{ if ((SIR_Error = Symbol->WriteSC(File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (10b) class instantiations", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

#ifndef NDEBUG
Symbol = First();
while(Symbol)
   { assert(Symbol->Color == SIR_WHITE);	/* assure clean start */
     Symbol = Symbol->Succ();
    } /* elihw */
#endif /* NDEBUG */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsBehaviorInstance())
	 ||(Symbol->IsChannelInstance()))
	{ if ((SIR_Error = WriteOrderedSC1(Symbol, File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsBehaviorInstance())
	 ||(Symbol->IsChannelInstance()))
	{ assert(Symbol->Color == SIR_RED);	/* assure work is done */
	  Symbol->Color = SIR_WHITE;
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (11) Interface definitions */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (11) interface definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsInterfaceDefinition())
	{ if ((SIR_Error = Symbol->WriteSC2(File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (12) Behavior and channel definitions (SORTED) */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (12) behavior and channel definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

#ifndef NDEBUG
Symbol = First();
while(Symbol)
   { assert(Symbol->Color == SIR_WHITE);	/* assure clean start */
     Symbol = Symbol->Succ();
    } /* elihw */
#endif /* NDEBUG */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsBehaviorDefinition())
	 ||(Symbol->IsChannelDefinition()))
	{ if ((SIR_Error = WriteOrderedSC2(Symbol, File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsBehaviorDefinition())
	 ||(Symbol->IsChannelDefinition()))
	{ assert(Symbol->Color == SIR_RED);	/* assure work is done */
	  Symbol->Color = SIR_WHITE;
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (13) Function/Method definitions */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (13) function/method definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsFunctionDefinition())
	{ if ((SIR_Error = Symbol->WriteSC3(File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::WriteSC */


ERROR SIR_Symbols::WriteOrderedSC1(	/* writes ordered instantiations */
	sir_symbol	*Current,	/* (recursively) */
	FILE		*File,
	BOOL		WriteNotes)
{
sir_symbol	*Symbol;

if (Current->Color == SIR_RED)		/* already written? */
   { return(SIR_ERROR_NO_ERROR);
    } /* fi */
if (Current->Color == SIR_YELLOW)	/* writing in progress? */
   { SIR_ErrMsg.form(
		"Cyclic dependency in class instantiation '%s' (internal)",
			Current->Name.chars());
     return(SIR_ERROR_INSTANCE_CYCLIC_NESTING);
    } /* fi */
Current->Color = SIR_YELLOW;	/* start working on this one */

Symbol = First();
while(Symbol)
   { if (Symbol->IsChannelInstance())
	{ if (Current->MapsPortTo(Symbol))
	     { if ((SIR_Error = WriteOrderedSC1(Symbol,
					File, WriteNotes)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

if ((SIR_Error = Current->WriteSC(File, WriteNotes)))
   { return(SIR_Error);
    } /* fi */
Current->Color = SIR_RED;	/* work is done for this one */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::WriteOrderedSC1 */


ERROR SIR_Symbols::WriteOrderedSC2(	/* writes ordered class definitions */
	sir_symbol	*Current,	/* (recursively) */
	FILE		*File,
	BOOL		WriteNotes)
{
sir_symbol	*Symbol;

if (Current->Color == SIR_RED)		/* already written? */
   { return(SIR_ERROR_NO_ERROR);
    } /* fi */
if (Current->Color == SIR_YELLOW)	/* writing in progress? */
   { SIR_ErrMsg.form(
		"Cyclic nesting in class definition '%s' (internal)",
			Current->Name.chars());
     return(SIR_ERROR_CLASS_CYCLIC_NESTING);
    } /* fi */
Current->Color = SIR_YELLOW;	/* start working on this one */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsBehaviorDefinition())
	 ||(Symbol->IsChannelDefinition()))
	{ if (Current->Instantiates(Symbol))
	     { if ((SIR_Error = WriteOrderedSC2(Symbol,
					File, WriteNotes)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

if ((SIR_Error = Current->WriteSC2(File, WriteNotes)))
   { return(SIR_Error);
    } /* fi */
Current->Color = SIR_RED;	/* work is done for this one */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::WriteOrderedSC2 */


ERROR SIR_Symbols::WriteH(	/* generates C++ header code */
	FILE		*File,
	BOOL		WriteNotes)
{
sir_symbol	*Symbol;

/* (1)-(4) User-defined types, declarations and definitions */

if (UserTypes)
   { if ((SIR_Error = UserTypes->WriteSC(File, WriteNotes, TRUE)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */

/* (5) Type definitions (alias types) */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (5) type definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsTypeDef())
	 &&((  (Symbol->Type->UserType == NULL)	/* except unnamed usertypes */
	     ||(Symbol->Type->UserType->Name != NULL))))
	{ if ((SIR_Error = Symbol->WriteSC(File, WriteNotes, FALSE, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (6) Class declarations */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (6) class declarations", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsClass())
	{ if ((SIR_Error = Symbol->WriteCC2(File, WriteNotes, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (7) Interface class definitions */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (7) interface class definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsInterfaceDefinition())
	{ if ((SIR_Error = Symbol->WriteCC2(File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (8) IP class definitions */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (8) IP class definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (  (  (Symbol->IsBehavior())	/* behavior and channel */
	    ||(Symbol->IsChannel()))	/* _declarations_ only  */
	 &&(! Symbol->ClassScope))
	{ if ((SIR_Error = Symbol->WriteCC2IP(File)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (9) Behavior and channel class definitions (SORTED) */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (9) behavior and channel class definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

#ifndef NDEBUG
Symbol = First();
while(Symbol)
   { assert(Symbol->Color == SIR_WHITE);	/* assure clean start */
     Symbol = Symbol->Succ();
    } /* elihw */
#endif /* NDEBUG */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsBehaviorDefinition())
	 ||(Symbol->IsChannelDefinition()))
	{ if ((SIR_Error = WriteOrderedCC2(Symbol, File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsBehaviorDefinition())
	 ||(Symbol->IsChannelDefinition()))
	{ assert(Symbol->Color == SIR_RED);	/* assure work is done */
	  Symbol->Color = SIR_WHITE;
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (10) Declarations of variables and functions */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (10) declarations of variables and functions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsVariable())
	 ||(Symbol->IsFunction()))
	{ if ((SIR_Error = Symbol->WriteSC(File, WriteNotes, TRUE, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::WriteH */


ERROR SIR_Symbols::WriteCC(	/* generates C++ main code */
	FILE		*File,
	BOOL		WriteNotes)
{
sir_symbol	*Symbol;

/* (1) Variable definitions */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (1) variable definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsVariableDefinition())
	{ if ((SIR_Error = Symbol->WriteSC(File, WriteNotes, FALSE, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (2) Channel implementations */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (2) channel implementations", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsChannelDefinition())
	{ if ((SIR_Error = Symbol->WriteCC2b(File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (3) Behavior implementations */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (3) behavior implementations", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsBehaviorDefinition())
	{ if ((SIR_Error = Symbol->WriteCC2b(File, WriteNotes)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

/* (4) Function definitions */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (4) function definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

Symbol = First();
while(Symbol)
   { if (Symbol->IsFunctionDefinition())
	{ if ((SIR_Error = Symbol->WriteSC3(File, WriteNotes, TRUE)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::WriteCC */


ERROR SIR_Symbols::WriteOrderedCC2(	/* writes ordered C++ class defs. */
	sir_symbol	*Current,	/* (recursively) */
	FILE		*File,
	BOOL		WriteNotes)
{
sir_symbol	*Symbol;

if (Current->Color == SIR_RED)		/* already written? */
   { return(SIR_ERROR_NO_ERROR);
    } /* fi */
if (Current->Color == SIR_YELLOW)	/* writing in progress? */
   { SIR_ErrMsg.form("Cyclic nesting in class definition '%s' (internal)",
			Current->Name.chars());
     return(SIR_ERROR_CLASS_CYCLIC_NESTING);
    } /* fi */
Current->Color = SIR_YELLOW;	/* start working on this one */

Symbol = First();
while(Symbol)
   { if (  (Symbol->IsBehaviorDefinition())
	 ||(Symbol->IsChannelDefinition()))
	{ if (Current->Instantiates(Symbol))
	     { if ((SIR_Error = WriteOrderedCC2(Symbol,
					File, WriteNotes)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	 } /* fi */
     Symbol = Symbol->Succ();
    } /* elihw */

if ((SIR_Error = Current->WriteCC2(File, WriteNotes)))
   { return(SIR_Error);
    } /* fi */
Current->Color = SIR_RED;	/* work is done for this one */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_Symbols::WriteOrderedCC2 */


	/*********************/
	/*** SIR_SymbolPtr ***/
	/*********************/


//++++++++++++++++++++++++++++ API Layer 1 +++++++++++++++++++++++++++++//


SIR_SymbolPtr::SIR_SymbolPtr(			/* constructor #1 */
	sir_symbol	*Symbol)
{

assert(Symbol != NULL);

SIR_SymbolPtr::Symbol	= Symbol;

} /* end of SIR_SymbolPtr::SIR_SymbolPtr #1 */


SIR_SymbolPtr::SIR_SymbolPtr(		/* constructor #3 (duplicator) */
	sir_symbol_ptr	*Original)
{

SIR_SymbolPtr::Symbol	= Original->Symbol;

} /* end of SIR_SymbolPtr::SIR_SymbolPtr #3 */


SIR_SymbolPtr::~SIR_SymbolPtr(void)		/* destructor */
{

/* nothing to do */

} /* end of SIR_SymbolPtr::~SIR_SymbolPtr */


	/**********************/
	/*** SIR_SymbolPtrs ***/
	/**********************/


//++++++++++++++++++++++++++++ API Layer 1 +++++++++++++++++++++++++++++//


SIR_SymbolPtrs::SIR_SymbolPtrs(		/* constructor #1 */
	sir_symbol_ptr	*FirstEntry = NULL) :
		SIR_List<SIR_SymbolPtr>(FirstEntry)
{

/* nothing else to do */

} /* end of SIR_SymbolPtrs::SIR_SymbolPtrs #1 */


SIR_SymbolPtrs::SIR_SymbolPtrs(		/* constructor #2 (duplicator) */
	sir_symbol_ptrs	*Original)
{
sir_symbol_ptr	*SymbolPtr;

assert(Original != NULL);

SymbolPtr = Original->First();
while(SymbolPtr)
   { Append(new SIR_SymbolPtr(SymbolPtr));
     SymbolPtr = SymbolPtr->Succ();
    } /* elihw */

} /* end of SIR_SymbolPtrs::SIR_SymbolPtrs #2 */


SIR_SymbolPtrs::~SIR_SymbolPtrs(void)	/* destructor */
{

/* nothing to do */

} /* end of SIR_SymbolPtrs::~SIR_SymbolPtrs */


void SIR_SymbolPtrs::UnAlias(void)/* unalias all type, usertype, symbol links */
{
sir_symbol_ptr	*SymbolPtr;

SymbolPtr = First();
while(SymbolPtr)
   { if (  (SymbolPtr->Symbol)
	 &&(SymbolPtr->Symbol->Alias))
	{ SymbolPtr->Symbol = SymbolPtr->Symbol->Alias;
	 } /* fi */
     SymbolPtr = SymbolPtr->Succ();
    } /* elihw */

} /* end of SIR_SymbolPtrs::UnAlias */


int SIR_SymbolPtrs::CmpSymbolNames(		/* compare all symbol names */
	sir_symbol_ptrs	*Others)		/* (similar to strcmp) */
{
sir_symbol_ptr	*ThisSymbPtr;
sir_symbol_ptr	*OtherSymbPtr;
int		CmpVal;

ThisSymbPtr = First();
OtherSymbPtr = Others->First();
while(ThisSymbPtr)
   { if (!(OtherSymbPtr))
	{ return(1);	/* this list has more entries */
	 } /* fi */
     if ((CmpVal = strcmp(ThisSymbPtr->Symbol->Name.chars(),
			OtherSymbPtr->Symbol->Name.chars())))
	{ return(CmpVal);
	 } /* fi */
     ThisSymbPtr = ThisSymbPtr->Succ();
     OtherSymbPtr = OtherSymbPtr->Succ();
    } /* elihw */
if (OtherSymbPtr)
   { return(-1);	/* this list has less entries */
    } /* fi */

return(0);	/* the lists are equal */

} /* end of SIR_SymbolPtrs::CmpSymbolNames */


sir_symbol_ptr *SIR_SymbolPtrs::Find(	/* searches for a specific entry */
	sir_symbol	*Symbol)	/* (returns NULL if not found) */
{
sir_symbol_ptr	*Ptr;

Ptr = First();
while(Ptr)
   { if (Ptr->Symbol == Symbol)
	{ break;	/* found! */
	 } /* fi */
     Ptr = Ptr->Succ();
    } /* elihw */

return(Ptr);

} /* end of SIR_SymbolPtrs::Find */


/*** internal functions *************************************************/


static void MarkImplementedMethods(	/* mark all methods with this name */
	const char	*Name,
	sir_symbol_ptr	*InterfaceList)
{
sir_symbol_ptr	*Interface;
sir_symbol	*Method;

Interface = InterfaceList;
while(Interface)
   { if (Interface->Symbol->ClassScope)
	{ Method = Interface->Symbol->ClassScope->First();
	  while(Method)
	     { if (0 == strcmp(Method->Name.chars(), Name))
		  { Method->Color = SIR_RED;	/* put marker */
		    break;
		   } /* fi */
	       Method = Method->Succ();
	      } /* elihw */
	 } /* fi */
     Interface = Interface->Succ();
    } /* elihw */

} /* end of MarkImplementedMethods */


/* EOF Symbol.cc */
