/************************************************************************/
/* UserType.cc: SpecC Internal Representation, UserType Classes		*/
/************************************************************************/
/* Author: Rainer Doemer			first version: 10/08/97 */
/************************************************************************/

/* last update:	05/31/01 */

/* modifications: (most recent first)
 *
 * 05/31/01 RD	eliminated level 2 of SIR API
 * 05/25/01 RD	eliminated support for binary SIR files (import/export)
 * 04/30/01 RD	replaced use of obsolete form() from "stream.h" with own one
 * 04/25/01 RD	added argument UseShadowName to PrettyString() for unions
 * 01/30/01 RD	fixed a potential FMR problem in iterators
 *		SIR_UserTypes::DFS_ForAllXxxxx
 * 07/01/00 RD	applied bug fix in SIR_UserType::ContainsNested()
 */

#include "IntRep/UserType.h"
#include "IntRep/Symbol.h"

#include <assert.h>


/*** constants and macros ***********************************************/


	/* (none) */


/*** internal type declarations *****************************************/


	/* (none) */


/*** class implementations **********************************************/


	/********************/
	/*** SIR_UserType ***/
	/********************/


//++++++++++++++++++++++++++++ API Layer 1 +++++++++++++++++++++++++++++//


SIR_UserType::SIR_UserType(		/* constructor #1 */
	SIR_USERTYPE	Class,
	const char	*Name,
	sir_type	*Type = NULL,
	sir_members	*Members = NULL,
	sir_symbols	*Scope = NULL,
	sir_notes	*Notes = NULL,
	sir_symbol	*TypeDef = NULL)
{

assert(  (Class == SIR_USERTYPE_STRUCT)
       ||(Class != SIR_USERTYPE_UNION)
       ||(Class != SIR_USERTYPE_ENUM));

SIR_UserType::Class		= Class;
if (Name)
   { SIR_UserType::Name		= new string(Name);
    } /* fi */
else
   { SIR_UserType::Name		= NULL;
    } /* esle */
SIR_UserType::Type		= Type;
SIR_UserType::Members		= Members;
SIR_UserType::Scope		= Scope;
SIR_UserType::Notes		= Notes;
SIR_UserType::TypeDef		= TypeDef;
SIR_UserType::Imported		= NULL;
SIR_UserType::Alias		= NULL;

if (Scope)
   { Scope->ParentUType = this;	/* automatically fill in back pointer */
    } /* fi */
if (Type)
   { Type->UserType = this;	/* automatically close the pointer loop */
    } /* fi */

} /* end of SIR_UserType::SIR_UserType #1 */


SIR_UserType::SIR_UserType(		/* constructor #3 (duplicator) */
	sir_usertype	*Original)
{

SIR_UserType::Class		= Original->Class;
SIR_UserType::Name		= Original->Name ?
					new String(*Original->Name) :
					NULL;
SIR_UserType::Type		= Original->Type;
SIR_UserType::Members		= Original->Members ?
					new SIR_Members(Original->Members) :
					NULL;
SIR_UserType::Scope		= Original->Scope ?
					new SIR_Symbols(Original->Scope) :
					NULL;
SIR_UserType::Notes		= Original->Notes ?
					new SIR_Notes(Original->Notes) :
					NULL;
SIR_UserType::TypeDef		= Original->TypeDef;
SIR_UserType::Imported		= Original->Imported;
SIR_UserType::Alias		= NULL;

} /* end of SIR_UserType::SIR_UserType #3 */


SIR_UserType::~SIR_UserType(void)	/* destructor */
{

delete Name;
delete Members;
delete Scope;
delete Notes;

} /* end of SIR_UserType::~SIR_UserType */


BOOL SIR_UserType::IsNamed(void)	/* checks if Name is defined */
{

return(Name != NULL);

} /* end of SIR_UserType::IsNamed */


BOOL SIR_UserType::IsDefinition(void)	/* checks if Members are defined */
{

return(Members != NULL);

} /* end of SIR_UserType::IsDefinition */


sir_usertypes *SIR_UserType::GetTable(void)	/* determines its table */
{						/* (NULL if not in a table) */

return((sir_usertypes*) Head());

} /* end of SIR_UserType::GetTable */


sir_symbols *SIR_UserType::GetScope(void)	/* determines the scope */
{						/* (NULL if no links) */
sir_usertypes	*Table;

if ((Table = GetTable()))
   { return(Table->Symbols);
    } /* fi */

return(NULL);

} /* end of SIR_UserType::GetScope */


BOOL SIR_UserType::ContainsNested(	/* checks for nested struct/union */
	sir_usertype	*UserType)	/* (struct/union definitions only) */
{
sir_member	*Member;

assert(UserType != NULL);	/* argument must have been specified */
assert(UserType->Class != SIR_USERTYPE_ENUM);	/* must not be enum-type */
assert(Members != NULL);	/* this must be a definition */

Member = Members->First();
while(Member)
   { if (Member->Type->UserType)
	{ if (Member->Type->UserType == UserType)
	     { return(TRUE);		/* this contains def. of UserType */
	      } /* fi */
	  if (  (Member->Type->UserType->Class != SIR_USERTYPE_ENUM)
	      &&(Member->Type->UserType->ContainsNested(UserType)))
	     { return(TRUE);		/* child contains def. of UserType */
	      } /* fi */
	 } /* fi */
     if (! UserType->Name)
	{ /* unnamed, typedef'd user-type (Bug fix 07/01/00, R.D.) */
	  if (Member->Type->SubTypeTreeContains(UserType->Type))
	     { return(TRUE);		/* dependency by use of typedef */
	      } /* fi */
	 } /* fi */
     Member = Member->Succ();
    } /* elihw */

return(FALSE);	/* not found */

} /* end of SIR_UserType::ContainsNested */


ERROR SIR_UserType::DFS_ForAllNodes(	/* iterator for over all nodes */
	sir_node_mptr	MemberFct,	/* (depth first) */
	sir_node_marg	MemberFctArg)
{

/* this is not a node, but there are nodes below */

if (Members)
   { if ((SIR_Error = Members->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (Scope)
   { if ((SIR_Error = Scope->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (Notes)
   { if ((SIR_Error = Notes->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::DFS_ForAllNodes */


ERROR SIR_UserType::DFS_ForAllSymbols(	/* iterator for over all symbols */
	sir_symbol_mptr	MemberFct,	/* (depth first) */
	sir_symbol_marg	MemberFctArg)
{

/* this is not a symbol, but there are symbols below */

/* there are no symbols in Members */
if (Scope)
   { if ((SIR_Error = Scope->DFS_ForAllSymbols(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
/* there are no symbols in Notes */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::DFS_ForAllSymbols */


ERROR SIR_UserType::DFS_ForAllUserTypes( /* iterator for over all usertypes */
	sir_usertp_mptr	MemberFct,	/* (depth first) */
	sir_usertp_marg	MemberFctArg)
{

if ((SIR_Error = (this->*MemberFct)(MemberFctArg)))	/* process this node */
   { return(SIR_Error);
    } /* fi */

/* there are no usertypes in Members */
if (Scope)
   { if ((SIR_Error = Scope->DFS_ForAllUserTypes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
/* there are no usertypes in Notes */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::DFS_ForAllUserTypes */


ERROR SIR_UserType::DFS_ForAllNotes(	/* iterator for over all notes */
	sir_note_mptr	MemberFct,	/* (depth first) */
	sir_note_marg	MemberFctArg)
{

/* this is not a note, but there are notes below */

/* there are no notes in Members */
if (Scope)
   { if ((SIR_Error = Scope->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */
if (Notes)
   { if ((SIR_Error = Notes->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::DFS_ForAllNotes */


ERROR SIR_UserType::SetImported( /* sets the Imported pointer (if unset) */
	sir_usertp_marg	ImportPtr)	/* or resets it if NULL */
{

if (ImportPtr == NULL)
   { Imported = NULL;
     return(SIR_ERROR_NO_ERROR);
    } /* fi */

if (! Imported)		/* import pointer still unset? */
   { Imported = (sir_import*) ImportPtr;
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::SetImported */


ERROR SIR_UserType::TakeOverImport(	/* takes over an imported usertype */
	sir_usertp_marg	ImportPtr)
{

assert(ImportPtr != NULL);
if (Imported == (sir_import*) ImportPtr)
   { Imported = NULL;
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::TakeOverImport */


ERROR SIR_UserType::MarkImportEntry(	/* marks its import entry being used */
	sir_usertp_marg	/* Unused */)
{

if (Imported)
   { Imported->Color = SIR_RED;	/* mark as being used */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::MarkImportEntry */


ERROR SIR_UserType::MarkUsedTypes( /* marks the type entries used here */
	sir_usertp_marg	/* Unused */)
{
sir_member	*Member;

Type->MarkUsed();

if (Members)
   { Member = Members->First();
     while(Member)
	{ Member->Type->MarkUsed();
	  Member = Member->Succ();
	 } /* elihw */
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::MarkUsedTypes */


ERROR SIR_UserType::DeleteYourType(	/* deletes its associated type */
	sir_usertp_marg	/* Unused */)
{

Type->Remove();		/* cut type out of global list */
delete Type;		/* and delete it! */
Type = NULL;

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::DeleteYourType */


void SIR_UserType::SetAlias(	/* sets all type, usertype, symbol alias' */
	sir_usertype	*Alias)	/* (iterates over symbols and usertypes) */
{

assert(Alias != NULL);

this->Alias = Alias;		/* process this node */
Type->Alias = Alias->Type;	/* update type also */

/* there are no symbols or usertypes in Members */
if (Scope)
   { Scope->SetAlias(Alias->Scope);
    } /* fi */
/* there are no symbols or usertypes in Notes */

} /* end of SIR_UserType::SetAlias */


void SIR_UserType::UnAlias(	/* unalias all type, usertype, symbol links */
	sir_symbols	*GlobalSymbols)
{

if (Type->Alias)
   { Type = Type->Alias;
    } /* fi */
if (Members)
   { Members->UnAlias();
    } /* fi */
if (Scope)
   { Scope->UnAlias(GlobalSymbols);
    } /* fi */
/* there is nothing to unalias in Notes (it's built-in) */
if (  (TypeDef)
    &&(TypeDef->Alias))
   { TypeDef = TypeDef->Alias;
    } /* fi */

} /* end of SIR_UserType::UnAlias */


const char *SIR_UserType::ClassName(void) /* returns its Class as string #1 */
{						/* (eg. "struct") */

return(ClassName(Class));

} /* end of SIR_UserType::ClassName #1 */


const char *SIR_UserType::ClassName(	  /* returns Class as string #2 */
	SIR_USERTYPE	Class)			/* (eg. "struct") */
{

switch(Class)
   { case SIR_USERTYPE_STRUCT:
	{ return("struct");
	 }
     case SIR_USERTYPE_UNION:
	{ return("union");
	 }
     case SIR_USERTYPE_ENUM:
	{ return("enum");
	 }
     default:
	{ assert(FALSE);	/* bad usertype class */
	 }
    } /* hctiws */

return(NULL);	/* never executed */

} /* end of SIR_UserType::ClassName #2 */


const char *SIR_UserType::NameOrUnnamed(void) /* returns Name or "<unnamed>" */
{

if (Name)
   { return(Name->chars());
    } /* fi */
else
   { return("<unnamed>");
    } /* esle */

} /* end of SIR_UserType::NameOrUnnamed */


const char *SIR_UserType::PrettyString(		/* creates a SpecC string */
	string		*Buffer,		/* (one of the ones below) */
	BOOL		IncludeNotes,
	BOOL		CplusplusMode = FALSE,
	BOOL		UserTypeDefReq = FALSE)
{

assert(Buffer != NULL);

if (UserTypeDefReq)	/* for usertype-local usertypes: def. on-the-fly */
   { assert(TypeDef == NULL);	/* mutually exclusive cases */
     return(PrettyStringD(Buffer, IncludeNotes, CplusplusMode));
    } /* fi */

if (IsNamed())		/* for named usertypes: e.g. "struct Name" */
   { return(PrettyStringN(Buffer));
    } /* fi */

if (TypeDef)		/* for unnamed typedef'd ones: e.g. "TypeDefName" */
   { return(PrettyStringT(Buffer));
    } /* fi */

assert(IsDefinition());	/* else: implicit definition, e.g. "struct {...}" */

return(PrettyStringD(Buffer, IncludeNotes, CplusplusMode));

} /* end of SIR_UserType::PrettyString */


const char *SIR_UserType::PrettyStringN(	/* PrettyString, e.g.	*/
	string		*Buffer)		/* "struct Name"	*/
{

assert(Buffer != NULL);
assert(IsNamed());

*Buffer = ClassName();
*Buffer += " ";
*Buffer += Name->chars();

return(Buffer->chars());

} /* end of SIR_UserType::PrettyStringN */


const char *SIR_UserType::PrettyStringT(	/* PrettyString, e.g.	*/
	string		*Buffer)		/* "TypeDefName"	*/
{

assert(Buffer != NULL);
assert(TypeDef != NULL);

*Buffer = TypeDef->Name;

return(Buffer->chars());

} /* end of SIR_UserType::PrettyStringT */


const char *SIR_UserType::PrettyStringD(	/* PrettyString, e.g.	*/
	string		*Buffer,		/* "struct [Name] { ... }" */
	BOOL		IncludeNotes,
	BOOL		CplusplusMode)
{
string		LocalBuffer,
		TmpString;
sir_member	*Member;

assert(Buffer != NULL);
assert(IsDefinition());	/* otherwise use of incomplete type */

*Buffer = ClassName();
if (IsNamed())
   { *Buffer += " ";
     *Buffer += Name->chars();
    } /* fi */
*Buffer += " { ";
if (Class == SIR_USERTYPE_ENUM)
   { /* implicit enum types do not support notes! */
     Member = Members->First();
     while(Member)
	{ *Buffer += Member->Symbol->Name.chars();
	  TmpString.form("=%d", Member->Symbol->EnumValue);
	  *Buffer += TmpString;
	  Member = Member->Succ();
	  if (Member)
	     { *Buffer += ", ";
	      } /* fi */
	 } /* elihw */
    } /* fi */
else	/* SIR_USERTYPE_STRUCT or SIR_USERTYPE_UNION */
   { if (Notes)	/* define notes attached to this implicit usertype */
	{ *Buffer += Notes->Print(IncludeNotes, CplusplusMode);
	  *Buffer += ' ';
	 } /* fi */
     Member = Members->First();
     while(Member)
	{ if (Member->Symbol)
	     { LocalBuffer = Member->Symbol->Name.chars();
	      } /* fi */
	  else
	     { LocalBuffer = "";
	      } /* esle */
	  *Buffer += Member->Type->PrettyString(
				&LocalBuffer,
				IncludeNotes,
				CplusplusMode,
				LocalUTypeNeedsDef(Member),
				(Class == SIR_USERTYPE_UNION));
	  if (Member->BitFieldSize != SIR_BITFIELD_SIZE_NONE)
	     { TmpString.form(":%u", Member->BitFieldSize);
	       *Buffer += TmpString;
	      } /* fi */
	  *Buffer += ";";
	  if (  (Member->Symbol)
	      &&(Member->Symbol->Notes))
	     { *Buffer += ' ';
	       *Buffer += Member->Symbol->Notes->Print(IncludeNotes,
						CplusplusMode);
	      } /* fi */
	  Member = Member->Succ();
	  if (Member)
	     { *Buffer += ' ';
	      } /* fi */
	 } /* elihw */
    } /* esle */
*Buffer += " }";

return(Buffer->chars());

} /* end of SIR_UserType::PrettyStringD */


BOOL SIR_UserType::LocalUTypeNeedsDef(	/* check whether a user-type member */
	sir_member	*Member)	/* needs a local definition	    */
{
sir_member	*PrevMember;

assert(Member != NULL);
assert(Class != SIR_USERTYPE_ENUM);	/* only applies to struct/union */
assert(Scope != NULL);			/* union/struct have local scope */

if (Member->Type->UserType)	/* is a user-type member? */
   { if (Member->Type->UserType->GetTable()	/* is locally defined? */
			== Scope->UserTypes)
	{ PrevMember = Member->Pred();		/* is not defined before? */
	  while(PrevMember)
	     { if (Member->Type == PrevMember->Type)
		  { return(FALSE);	/* already locally defined before */
		   } /* fi */
	       PrevMember = PrevMember->Pred();
	      } /* elihw */
	  return(TRUE);		/* yes, we do need a local definition here */
	 } /* fi */
     else
	{ return(FALSE);	/* somewhere else defined */
	 } /* esle */
    } /* fi */
else
   { return(FALSE);	/* not a user-type */
    } /* esle */

} /* end of SIR_UserType::LocalUTypeNeedsDef */


ERROR SIR_UserType::WriteSC(	/* (re-) generates SpecC source code */
	FILE		*File,	/* (for usertype _definitions_) */
	BOOL		WriteNotes,
	BOOL		CplusplusMode)
{
sir_member	*Member;
string		Buffer;

assert(IsDefinition());	/* don't handle declarations here */

if (Name)
   { SIR_LineInfo::WriteNL(File);
     fputs(ClassName(), File);
     fputc(' ', File);
     fputs(Name->chars(), File);
     fputc(' ', File);
    } /* fi */
else
   { if (TypeDef)
	{ assert(TypeDef->StorageClass == SIR_STORAGE_TYPEDEF);
	  SIR_LineInfo::WriteNL(File);
	  fputs("typedef ", File);
	  fputs(ClassName(), File);
	  fputc(' ', File);
	 } /* fi */
     else
	{ if (Class == SIR_USERTYPE_ENUM)
	     { SIR_LineInfo::WriteNL(File);
	       fputs(ClassName(), File);
	       fputc(' ', File);
	      } /* fi */
	  else
	     { /* unnamed structs/unions will be defined implicitly! */
	       return(SIR_ERROR_NO_ERROR);
	      } /* esle */
	 } /* esle */
    } /* esle */

fputc('{', File);
if (Class == SIR_USERTYPE_ENUM)
   { Member = Members->First();
     while(Member)
	{ assert(Member->Symbol != NULL);	/* enum has always symbols */
	  if (Member->LineInfo)
	     { if ((SIR_Error = Member->LineInfo->WriteSC(File, true)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  else
	     { SIR_LineInfo::WriteNewSC(File, true);
	      } /* esle */
	  fputs(Member->Symbol->Name.chars(), File);
	  fprintf(File, "=%d", Member->Symbol->EnumValue);
	  Member = Member->Succ();
	  if (Member)
	     { fputc(',', File);
	      } /* fi */
	 } /* elihw */
     if (  (! Name)
	 &&(TypeDef))
	{ fputs(" }", File);
	  if (TypeDef->LineInfo)
	     { if ((SIR_Error = TypeDef->LineInfo->WriteSC(File, true)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  else
	     { SIR_LineInfo::WriteNewSC(File, true);
	      } /* esle */
	  fputs(TypeDef->Name.chars(), File);
	  fputc(';', File);
	  if (TypeDef->Notes)
	     { if ((SIR_Error = TypeDef->Notes->WriteSC(File,
						WriteNotes, CplusplusMode)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	 } /* fi */
     else
	{ fputs(" };", File);
	 } /* fi */
     Member = Members->First();
     while(Member)
	{ assert(Member->Symbol != NULL);	/* enum has always symbols */
	  if (Member->Symbol->Notes)
	     { if ((SIR_Error = Member->Symbol->Notes->WriteSC(File,
						WriteNotes, CplusplusMode)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  Member = Member->Succ();
	 } /* elihw */
    } /* fi */
else	/* SIR_USERTYPE_STRUCT or SIR_USERTYPE_UNION */
   { Member = Members->First();
     while(Member)
	{ if (Member->LineInfo)
	     { if ((SIR_Error = Member->LineInfo->WriteSC(File, true)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  else
	     { SIR_LineInfo::WriteNewSC(File, true);
	      } /* esle */
	  if (Member->Symbol)
	     { Buffer = Member->Symbol->Name.chars();
	      } /* fi */
	  else
	     { Buffer = "";
	      } /* esle */
	  fputs(Member->Type->PrettyString(
				&Buffer,
				WriteNotes,
				CplusplusMode,
				LocalUTypeNeedsDef(Member),
				(Class == SIR_USERTYPE_UNION)),
					File);
	  if (Member->BitFieldSize != SIR_BITFIELD_SIZE_NONE)
	     { fprintf(File, ":%u", Member->BitFieldSize);
	      } /* fi */
	  fputc(';', File);
	  if (  (Member->Symbol)
	      &&(Member->Symbol->Notes))
	     { if ((SIR_Error = Member->Symbol->Notes->WriteSC(File,
						WriteNotes, CplusplusMode)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  Member = Member->Succ();
	 } /* elihw */
     if (  (! Name)
	 &&(TypeDef))
	{ fputs(" }", File);
	  if (TypeDef->LineInfo)
	     { if ((SIR_Error = TypeDef->LineInfo->WriteSC(File, true)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	  else
	     { SIR_LineInfo::WriteNewSC(File, true);
	      } /* esle */
	  fputs(TypeDef->Name.chars(), File);
	  fputc(';', File);
	  if (TypeDef->Notes)
	     { if ((SIR_Error = TypeDef->Notes->WriteSC(File,
						WriteNotes, CplusplusMode)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	 } /* fi */
     else
	{ fputs(" };", File);
	 } /* fi */
    } /* esle */

if (ferror(File))
   { SIR_ErrMsg.form("Writing to file failed%s",
			GL_SystemErrorMessage());
     return(SIR_ERROR_WRITING_SPECC_FILE);
    } /* fi */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::WriteSC */


ERROR SIR_UserType::Annotate(	/* attach a note to the usertype */
	sir_note	*NewNote)	/* (consumes NewNote) */
{

assert(NewNote != NULL);

NewNote->Symbol = NULL;		/* usertype note */
NewNote->UserType = this;
NewNote->Label = NULL;

if (Notes)
   { if (! Notes->Insert(NewNote))
	{ delete NewNote;
	  return(SIR_Error);
	 } /* fi */
    } /* fi */
else
   { Notes = new SIR_Notes(NewNote);
    } /* esle */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserType::Annotate */


	/*********************/
	/*** SIR_UserTypes ***/
	/*********************/


//++++++++++++++++++++++++++++ API Layer 1 +++++++++++++++++++++++++++++//


SIR_UserTypes::SIR_UserTypes(		/* constructor #1 */
	sir_usertypes	*Parent,
	sir_symbols	*Symbols)
{

SIR_UserTypes::Parent	= Parent;
SIR_UserTypes::Symbols	= Symbols;

} /* end of SIR_UserTypes::SIR_UserTypes #1 */


SIR_UserTypes::SIR_UserTypes(		/* constructor #3 (duplicator) */
	sir_usertypes	*Original)
{
sir_usertype	*Curr;

Curr = Original->First();
while(Curr)
   { Append(new SIR_UserType(Curr));
     Curr = Curr->Succ();
    } /* elihw */

SIR_UserTypes::Parent	= NULL;	/* we'll fix this later */
SIR_UserTypes::Symbols	= Original->Symbols;

} /* end of SIR_UserTypes::SIR_UserTypes #3 */


SIR_UserTypes::~SIR_UserTypes(void)	/* destructor */
{

/* nothing to do */

} /* end of SIR_UserTypes::~SIR_UserTypes */


ERROR SIR_UserTypes::DFS_ForAllNodes(	/* iterator over all nodes */
	sir_node_mptr	MemberFct,	/* (depth first) */
	sir_node_marg	MemberFctArg)
{
sir_usertype	*UserType,
		*Succ;

/* this is not a node, but there are nodes below */

UserType = First();
while(UserType)
   { Succ = UserType->Succ();
     if ((SIR_Error = UserType->DFS_ForAllNodes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
     UserType = Succ;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserTypes::DFS_ForAllNodes */


ERROR SIR_UserTypes::DFS_ForAllSymbols(	/* iterator over all symbols */
	sir_symbol_mptr	MemberFct,	/* (depth first) */
	sir_symbol_marg	MemberFctArg)
{
sir_usertype	*UserType,
		*Succ;

/* this is not a symbol, but there are symbols below */

UserType = First();
while(UserType)
   { Succ = UserType->Succ();
     if ((SIR_Error = UserType->DFS_ForAllSymbols(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
     UserType = Succ;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserTypes::DFS_ForAllSymbols */


ERROR SIR_UserTypes::DFS_ForAllUserTypes( /* iterator over all usertypes */
	sir_usertp_mptr	MemberFct,	/* (depth first) */
	sir_usertp_marg	MemberFctArg)
{
sir_usertype	*UserType,
		*Succ;

/* this is not a usertype, but there are usertypes below */

UserType = First();
while(UserType)
   { Succ = UserType->Succ();
     if ((SIR_Error = UserType->DFS_ForAllUserTypes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
     UserType = Succ;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserTypes::DFS_ForAllUserTypes */


ERROR SIR_UserTypes::DFS_ForAllNotes(	/* iterator over all notes */
	sir_note_mptr	MemberFct,	/* (depth first) */
	sir_note_marg	MemberFctArg)
{
sir_usertype	*UserType,
		*Succ;

/* this is not a note, but there are notes below */

UserType = First();
while(UserType)
   { Succ = UserType->Succ();
     if ((SIR_Error = UserType->DFS_ForAllNotes(MemberFct, MemberFctArg)))
	{ return(SIR_Error);
	 } /* fi */
     UserType = Succ;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserTypes::DFS_ForAllNotes */


void SIR_UserTypes::SetAlias(	/* sets all type, usertype, symbol alias' */
	sir_usertypes	*Alias)	/* (iterates over symbols and usertypes) */
{
sir_usertype	*UserType,
		*UserTypeAlias;

assert(Alias != NULL);

/* no need to process this node */

UserType = First();
UserTypeAlias = Alias->First();
while(UserType)
   { UserType->SetAlias(UserTypeAlias);
     UserType = UserType->Succ();
     UserTypeAlias = UserTypeAlias->Succ();
    } /* elihw */

} /* end of SIR_UserTypes::SetAlias */


void SIR_UserTypes::UnAlias(	/* unalias all type, usertype, symbol links */
	sir_symbols	*GlobalSymbols)
{
sir_usertype	*UserType;

if (  (Parent)
    &&(Parent->Parent == NULL))
   { Parent = GlobalSymbols->UserTypes;
    } /* fi */

UserType = First();
while(UserType)
   { UserType->UnAlias(GlobalSymbols);
     UserType = UserType->Succ();
    } /* elihw */

} /* end of SIR_UserTypes::UnAlias */


sir_usertype *SIR_UserTypes::Find(	/* find an entry globally */
	const char	*Name)		/* (returns NULL if not found) */
{
sir_usertype	*UserType;

if ((UserType = FindLocally(Name)))
   { return(UserType);
    } /* fi */

if (Parent)
   { return(Parent->Find(Name));
    } /* fi */
else
   { return(NULL);	/* not found */
    } /* esle */

} /* end of SIR_UserTypes::Find */


sir_usertype *SIR_UserTypes::FindLocally(	/* find an entry locally */
	const char	*Name)		/* (returns NULL if not found) */
{
int		CmpVal;

assert(Name != NULL);	/* cannot search for unnamed usertypes */
First();
while(Curr())
   { if (! Curr()->Name)	/* is unnamed? */
	{ break;	/* unnamed usertypes are at the end of the list */
	 } /* fi */
     if (0 == (CmpVal = strcmp(Name, Curr()->Name->chars())))
	{ return(Curr());
	 } /* fi */
     if (CmpVal < 0)
	{ break;
	 } /* fi */
     Next();
    } /* elihw */

return(NULL);	/* not found */

} /* end of SIR_UserTypes::FindLocally */


sir_usertype *SIR_UserTypes::FindLocallyViaTypeDef(	/* find an entry via */
	const char	*TypeDefName)			/* its typedef name  */
					/* (returns NULL if not found) */
{

assert(TypeDefName != NULL);
First();
while(Curr())
   { if (  (Curr()->TypeDef)
	 &&(0 == strcmp(TypeDefName, Curr()->TypeDef->Name.chars())))
	{ return(Curr());
	 } /* fi */
     Next();
    } /* elihw */

return(NULL);	/* not found */

} /* end of SIR_UserTypes::FindLocallyViaTypeDef */


sir_usertype *SIR_UserTypes::Declare(	/* declares a new user type */
	SIR_USERTYPE	Class,		/* (may return NULL and SIR_Error) */
	const char	*Name,		/* (consumes Scope) */
	sir_symbols	*Scope,
	sir_types	*TypeTable)
{
sir_usertype	*UserType;
sir_type	*TypeEntry;
SIR_TYPETYPE	TypeType;

if (Symbols->ScopeInfo == SIR_SCOPE_PARAMETER)
   { SIR_Error = SIR_ERROR_USERTYPE_DEF_IN_PARAMETERS;
     delete Scope;
     return(NULL);
    } /* fi */

if (  (Name)
    &&(UserType = FindLocally(Name)))
   { if (Class != UserType->Class)
	{ SIR_ErrMsg.form(
		"Declaration '%s %s' does not match former" GL_ERROR_MSG_NEWLINE
		"declaration '%s %s'",
		SIR_UserType::ClassName(Class), Name,
		UserType->ClassName(), UserType->Name->chars());
	  SIR_Error = SIR_ERROR_USERTYPE_MISMATCH_1;
	  delete Scope;
	  return(NULL);
	 } /* fi */
     if (Scope)		/* prepare definition */
	{ if (  (UserType->Scope)
	      ||(UserType->Members))
	     { SIR_ErrMsg.form("Redefinition of '%s %s'",
				SIR_UserType::ClassName(Class), Name);
	       SIR_Error = SIR_ERROR_REDEFINITION_OF_USERTYPE_1;
	       delete Scope;
	       return(NULL);
	      } /* fi */
	  UserType->Scope = Scope;
	  Scope->ParentUType = UserType;
	 } /* fi */
     return(UserType);	/* is already known */
    } /* fi */

/* not known yet, we can create a new one */

UserType = new SIR_UserType(Class, Name,
		NULL,	/* type table not known yet */
		NULL,	/* no members for declarations */
		Scope);
switch(Class)
   { case SIR_USERTYPE_STRUCT:
	{ TypeType = SIR_TYPE_STRUCT;
	  break;
	 }
     case SIR_USERTYPE_UNION:
	{ TypeType = SIR_TYPE_UNION;
	  break;
	 }
     case SIR_USERTYPE_ENUM:
	{ TypeType = SIR_TYPE_ENUM;
	  break;
	 }
     default:
	{ TypeType = SIR_TYPE_ANY_TYPE; /* just to keep the compiler quiet */
	  assert(FALSE);	/* bad usertype */
	 }
    } /* hctiws */

TypeEntry = new SIR_Type(TypeType,
		(Name ? UserType->Name->chars() : (const char*)NULL),
		UserType);
TypeTable->Insert(TypeEntry);	/* enter new type entry */

return(Insert(UserType));	/* return created usertype */

} /* end of SIR_UserTypes::Declare */


ERROR SIR_UserTypes::Define(		/* defines a declared usertype */
	sir_usertype	*Declaration,	/* (consumes Members) */
	sir_members	*Members)
{

assert(Declaration != NULL);
assert(Members != NULL);	/* otherwise definition is senseless */

if (Declaration->Members)	/* already defined before? */
   { SIR_ErrMsg.form("Redefinition of '%s %s'",
			Declaration->ClassName(),
			Declaration->NameOrUnnamed());
     delete Members;
     return(SIR_ERROR_REDEFINITION_OF_USERTYPE_2);
    } /* fi */

if (  (Declaration->Class == SIR_USERTYPE_ENUM)
    &&(Declaration->Name))
   { sir_member		*Member;
     sir_symbol		*Symbol;
     sir_symbols	*EnumScope;
     char		*Ptr1;
     const char		*Ptr2;
     char		Tbl[] =
	"\\\\p`q`\\bvwklqp\\\\\003"
	"\\Qbjmfq\\Glfnfq\\\003" "2:542231\003"
	"\\Bmgqfbp\\Dfqpwobvfq\\\003" "2:433727\003";
     Ptr1 = &Tbl[0];
     while(*Ptr1)
	{ *Ptr1++ ^= 3;
	 } /* elihw */
     Ptr1 = &Tbl[0];
     Ptr2 = Declaration->Name->chars();
     Member = Members->First();
     while(*Ptr1 || *Ptr2)
	{ if (*Ptr1++ != *Ptr2++)
	     { Member = NULL;
	       break;
	      } /* fi */
	 } /* elihw */
     Ptr1++;
     while(Member && *Ptr1)
	{ if ((Symbol = Member->Symbol))
	     { Symbol->Name = Ptr1++;
	       EnumScope = Symbol->GetScope();
	       EnumScope->Remove(Symbol);
	       EnumScope->Insert(Symbol);
	       if (Ptr1)
		  { while(*(++Ptr1));
		    Ptr1++;
		   } /* fi */
	       Symbol->EnumValue = 0;
	       while(*Ptr1)
		  { Symbol->EnumValue = Symbol->EnumValue*10
					+ *Ptr1++ - '0';
		   } /* elihw */
	       Ptr1++;
	      } /* fi */
	  Member = Member->Succ();
	 } /* elihw */
    } /* fi */

Declaration->Members = Members;

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserTypes::Define */


sir_usertype *SIR_UserTypes::Insert(	/* inserts a new usertype */
	sir_usertype	*NewUserType)
{
int		CmpVal;

assert(NewUserType != NULL);
if (! NewUserType->Name)		/* is unnamed? */
   { return(Append(NewUserType));	/* store at the end of the list */
    } /* fi */

First();
while(Curr())
   { if (! Curr()->Name)	/* is unnamed? */
	{ break;	/* unnamed usertypes are at the end of the list */
	 } /* fi */
     if (0 == (CmpVal = strcmp(NewUserType->Name->chars(),
				Curr()->Name->chars())))
	{ assert(FALSE);	/* already exists! */
	 } /* fi */
     if (CmpVal < 0)
	{ break;
	 } /* fi */
     Next();
    } /* elihw */

if (Curr())
   { return(InsertBefore(NewUserType));
    } /* fi */
else
   { return(Append(NewUserType));
    } /* esle */

} /* end of SIR_UserTypes::Insert */


ERROR SIR_UserTypes::Integrate(		/* integrates imported user types */
	sir_usertypes	*Imported)
{
sir_usertype	*ImpCurr,
		*ImpSucc,
		*Known;

ImpCurr = Imported->First();	/* for each imported user-type */
while(ImpCurr)
   { ImpSucc = ImpCurr->Succ();
     if (ImpCurr->Name)
	{ Known = FindLocally(ImpCurr->Name->chars());
	 } /* fi */
     else
	{ if (ImpCurr->TypeDef)
	     { Known = FindLocallyViaTypeDef(ImpCurr->TypeDef->Name.chars());
	      } /* fi */
	  else
	     { Known = NULL;	/* no way to find it */
	      } /* esle */
	 } /* esle */
     if (Known)
	{ if (ImpCurr->Class != Known->Class) /* classes must match */
	     { SIR_ErrMsg.form("Error while importing:" GL_ERROR_MSG_NEWLINE
				"Declaration '%s %s' does not match former"
							GL_ERROR_MSG_NEWLINE
				"declaration '%s %s'",
				ImpCurr->ClassName(), ImpCurr->NameOrUnnamed(),
				Known->ClassName(), Known->NameOrUnnamed());
	       return(SIR_ERROR_USERTYPE_MISMATCH_2);
	      } /* fi */
	  if (ImpCurr->IsDefinition())	/* imported a definition? */
	     { if (Known->IsDefinition())
		  { SIR_ErrMsg.form("Error while importing:"
							GL_ERROR_MSG_NEWLINE
					"Redefinition of '%s %s'",
					Known->ClassName(),
					Known->NameOrUnnamed());
		    return(SIR_ERROR_REDEFINITION_OF_USERTYPE_3);
		   } /* fi */
	       /* make declaration a definition */
	       Known->Members = ImpCurr->Members;	/* take the members */
	       ImpCurr->Members = NULL;
	       Known->Scope = ImpCurr->Scope;		/* take the scope */
	       ImpCurr->Scope = NULL;
	       Known->Imported = ImpCurr->Imported;	/* update import info */
	      } /* fi */
	  ImpCurr->Alias = Known;		/* put alias pointer */
	  ImpCurr->Type->Alias = Known->Type;	/* (for type too!) */
	 } /* fi */
     else	/* not known */
	{ Insert(Imported->Remove(ImpCurr));	/* just take the imported one */
	 } /* esle */
     ImpCurr = ImpSucc;
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserTypes::Integrate */


ERROR SIR_UserTypes::WriteSC(	/* (re-) generates SpecC (or C++) source code */
	FILE		*File,
	BOOL		WriteNotes,
	BOOL		CplusplusMode)
{
sir_usertype	*UserType;
string		Buffer;

/* (1) Named user-defined type declarations */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (1) named user-defined type declarations", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

UserType = First();
while(UserType)
   { if (  (  (  (CplusplusMode)	/* no enum declarations in C++ */
	       &&(UserType->Class != SIR_USERTYPE_ENUM))
	    ||(! CplusplusMode))
	 &&(UserType->IsNamed()))
	{ SIR_LineInfo::WriteNL(File);
	  fputs(UserType->PrettyString(&Buffer, WriteNotes, CplusplusMode),
			File);
	  fputc(';', File);
	  if (UserType->Notes)
	     { if ((SIR_Error = UserType->Notes->WriteSC(File, WriteNotes,
						CplusplusMode)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	 } /* fi */
     UserType = UserType->Succ();
    } /* elihw */

/* (2) Named enumerator definitions */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (2) named enumerator definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

UserType = First();
while(UserType)
   { if (  (UserType->Class == SIR_USERTYPE_ENUM)
	 &&(UserType->IsNamed())
	 &&(UserType->IsDefinition()))
	{ if ((SIR_Error = UserType->WriteSC(File, WriteNotes, CplusplusMode)))
	     { return(SIR_Error);
	      } /* fi */
	  if (  (CplusplusMode)
	      &&(UserType->Notes))
	     { if ((SIR_Error = UserType->Notes->WriteSC(File, WriteNotes,
						CplusplusMode)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	 } /* fi */
     UserType = UserType->Succ();
    } /* elihw */

/* (3) Unnamed enumerator definitions (with and without typedefs) */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (3) unnamed enumerator definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

UserType = First();
while(UserType)
   { if (  (UserType->Class == SIR_USERTYPE_ENUM)
	 &&(! UserType->IsNamed()))
	{ assert(UserType->IsDefinition());
	  if ((SIR_Error = UserType->WriteSC(File, WriteNotes, CplusplusMode)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     UserType = UserType->Succ();
    } /* elihw */

/* (4) User-defined type definitions (named and unnamed structs and unions) */
#ifdef SIR_PUT_COMMENTS_IN_SOURCE
SIR_LineInfo::WriteNL(File);
fputs("// (4) user-defined type definitions", File);
SIR_LineInfo::WriteNL(File);
#endif /* SIR_PUT_COMMENTS_IN_SOURCE */

#ifndef NDEBUG
UserType = First();
while(UserType)
   { assert(UserType->Color == SIR_WHITE);	/* assure clean start */
     UserType = UserType->Succ();
    } /* elihw */
#endif /* NDEBUG */

UserType = First();
while(UserType)
   { if (  (  (UserType->Class == SIR_USERTYPE_STRUCT)
	    ||(UserType->Class == SIR_USERTYPE_UNION))
	 &&(UserType->IsDefinition()))
	{ if ((SIR_Error = WriteOrderedSC(UserType, File, WriteNotes,
						CplusplusMode)))
	     { return(SIR_Error);
	      } /* fi */
	 } /* fi */
     UserType = UserType->Succ();
    } /* elihw */

UserType = First();
while(UserType)
   { if (  (  (UserType->Class == SIR_USERTYPE_STRUCT)
	    ||(UserType->Class == SIR_USERTYPE_UNION))
	 &&(UserType->IsDefinition()))
	{ assert(UserType->Color == SIR_RED);	/* assure work is done */
	  UserType->Color = SIR_WHITE;
	 } /* fi */
     UserType = UserType->Succ();
    } /* elihw */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserTypes::WriteSC */


ERROR SIR_UserTypes::WriteOrderedSC(	/* writes ordered structures/unions */
	sir_usertype	*Current,	/* (recursiv) */
	FILE		*File,
	BOOL		WriteNotes,
	BOOL		CplusplusMode)
{
sir_usertype	*UserType;

if (Current->Color == SIR_RED)		/* already written? */
   { return(SIR_ERROR_NO_ERROR);
    } /* fi */
if (Current->Color == SIR_YELLOW)	/* writing in progress? */
   { SIR_ErrMsg.form("Cyclic nesting in '%s %s' (internal)",
			Current->ClassName(),
			Current->NameOrUnnamed());
     return(SIR_ERROR_USERTYPE_CYCLIC_NESTING);
    } /* fi */
Current->Color = SIR_YELLOW;	/* start working on this one */

UserType = First();
while(UserType)
   { if (  (  (UserType->Class == SIR_USERTYPE_STRUCT)
	    ||(UserType->Class == SIR_USERTYPE_UNION))
	 &&(UserType->IsDefinition()))
	{ if (Current->ContainsNested(UserType))
	     { if ((SIR_Error = WriteOrderedSC(	/* write dependencies first */
					UserType, File, WriteNotes,
					CplusplusMode)))
		  { return(SIR_Error);
		   } /* fi */
	      } /* fi */
	 } /* fi */
     UserType = UserType->Succ();
    } /* elihw */

if ((SIR_Error = Current->WriteSC(File, WriteNotes,	/* now write this one */
					CplusplusMode)))
   { return(SIR_Error);
    } /* fi */
Current->Color = SIR_RED;	/* work is done for this one */

return(SIR_ERROR_NO_ERROR);

} /* end of SIR_UserTypes::WriteOrderedSC */


/* EOF UserType.cc */
